/*++

(C) Copyright 2000-2006 Hewlett-Packard Development Company, L.P.

Module Name:

    CpqCi.h

Abstract:

    This header file contains the support constructs necessary
    to implement the common components of the Compaq host channel 
    transport interface.

Environment:

    This module can be shared by kernel mode (drivers), user
    user mode (agents), and embedded software.  Compilers should
    be configured to pack data structures on 1-byte boundaries.

Compiler Settings:

    This module requires the following compiler settings.

        *   1-byte packed data structures
        *   All functions must adhere to standard 'C' calling convention
        

Revision History:

--*/

#ifndef _CPQCIh
#define _CPQCIh

#define CPQCIPRIVATE  static  //globals get scoped within a module

/******************************************************************************
                            *** BUGBUG ***
    The following constants control code generation.  These defines should be 
    commented out for a live build.

        CPQCI_DB_BITMAPPD  Generates code to have apps touch bit-mapped
                           doorbell register located at MMIOBAR+0x00D0.
                           Code is generated to touch the page-mapped
                           Doorbell addresses with this term undefined.

        CPQCI_SW_EMULATOR  Generates code to work with the IOP emulator
                           implemented as another host processor allocated
                           away from the host operating system.

******************************************************************************/

//#define CPQCI_DB_BITMAPPD  //bugbug: DO NOT DEFINE FOR LIVE BUILD
//#define CPQCI_SW_EMULATOR  //bugbug: DO NOT DEFINE FOR LIVE BUILD

/******************************************************************************
    The software emulator MUST have doorbells bit mapped.
******************************************************************************/

#ifdef  CPQCI_SW_EMULATOR
#ifndef CPQCI_DB_BITMAPPD
#define CPQCI_DB_BITMAPPD
#endif
#endif

/******************************************************************************
    The following stuff is not used by the embedded IOP environment.  The 
    current IOP tools (Greenhills) does not support 64-bit data-types.
******************************************************************************/

#ifndef CPQCI_EMBEDDED_IOP

#ifdef _MSC_EXTENSIONS
#define i64_t __int64   // compiler dependent!!!
#define __ccall__ _cdecl
#else
#ifdef CPQCI_X86_64
typedef long i64_t;
#else
typedef long long i64_t;
#endif
#define __ccall__
#endif

#ifdef CPQCI_X86_64
typedef unsigned long u64_t;
typedef volatile long vi64_t;
typedef volatile unsigned long vu64_t;
#else
typedef unsigned long long u64_t;
typedef volatile long long vi64_t;
typedef volatile unsigned long long vu64_t;
#endif

//*******************************************************************
//* Obtain a reference to the processor specific code to perform
//* atomic operations and create the prototype.
//*******************************************************************
#if defined(CPQCI_ASM) || defined(CPQCI_X86_64)
extern int CpqCiCas(vu64_t *Memory,u64_t Expected,u64_t New);
#else
extern unsigned char CpqCiCasCode[]; //located in the cpu specific module.
typedef int (__ccall__ *CPQCICASROUTINE)(vu64_t *Memory,u64_t Expected,u64_t New);
#define CpqCiCas ((CPQCICASROUTINE)(CpqCiCasCode))
#endif

#endif //CPQCI_EMBEDDED_IOP

/******************************************************************************
    Data types defined for the exported APIs.
******************************************************************************/

typedef void * CPQCIFIFOBAR;
typedef void * CPQCIDESCBAR;
typedef void * CPQCIMMIOBAR;
typedef void * CPQCIHANDLE;
typedef void * CPQCIBUFFER;
typedef void * CPQCIPACKET;
typedef void * CPQCIDOORBELL;

typedef u_t CPQCIPACKETID;
typedef u_t CPQCICONTROL;
typedef u_t CPQCILENGTH;
typedef u_t CPQCIENTRY;

typedef u8_t CPQCIBYTE;

#define CPQCINULL ((CPQCIVOID *)(0))

typedef enum _CPQCIFIFOTYPE {
    CpqCiSendQ,
    CpqCiRecvQ
} CPQCIFIFOTYPE;




#define CPQCI_NORMPRI_CHANNELS\
    ((CPQCI_BASE_HIGHPRIORITY_CHANNEL-CPQCI_BASE_NORMALPRIORITY_CHANNEL))

#define CPQCI_HIGHPRI_CHANNELS\
    ((CPQCI_BASE_CRITICALPRIORITY_CHANNEL-CPQCI_BASE_HIGHPRIORITY_CHANNEL))

#define CPQCI_CRI0PRI_CHANNELS\
    ((CpqCiCriticalPriority_0-CPQCI_BASE_CRITICALPRIORITY_CHANNEL) + 1)

#define CPQCI_CRI1PRI_CHANNELS\
    ((CpqCiCriticalPriority_1-CpqCiCriticalPriority_0))

#define CPQCI_CRI2PRI_CHANNELS\
    ((CpqCiCriticalPriority_2-CpqCiCriticalPriority_1))

#define CPQCI_CRI3PRI_CHANNELS\
    ((CpqCiCriticalPriority_3-CpqCiCriticalPriority_2))

#define CPQCI_CRI4PRI_CHANNELS\
    ((CpqCiCriticalPriority_4-CpqCiCriticalPriority_3))

#define CPQCI_CRI5PRI_CHANNELS\
    ((CpqCiCriticalPriority_5-CpqCiCriticalPriority_4))

#define CPQCI_CRI6PRI_CHANNELS\
    ((CpqCiCriticalPriority_6-CpqCiCriticalPriority_5))

#define CPQCI_CRI7PRI_CHANNELS\
    ((CpqCiCriticalPriority_7-CpqCiCriticalPriority_6))

/******************************************************************************
    Cacheline awareness is important for these algorithms to operate 
    efficiently.  The minimum line size will be targeted for Mckinley
    (128-byte cacheline).
******************************************************************************/

#define CPQCILINESIZE   128

/******************************************************************************
     CPQCI AMORPHOUS DATA TYPES OF FIXED SIZE.
******************************************************************************/

typedef union { u_t uint; u8_t *ptr; u8_t  _pad[CPQCILINESIZE];} CPQCITYPELINE;

/******************************************************************************
                        CPQCI PHYSICAL TRANSPORT CONSTRUCTS.
    Target systems will support a page granularity of at least 4k (except ALPHA
    systems which support 8K pages and not currently under consideration for 
    support).  There are currently 32-channels supported for the CPQCI device 
    with a doorbell page span of 16K-bytes per channel's doorbell region.
******************************************************************************/


#define CPQCI_MAX_DESCRIPTORS      (4*1024)  // 4k descriptors
#define CPQCI_MAX_DESCRIPTOR_SIZE  (4*1024)  // 4k bytes
#define CPQCI_MIN_DESCRIPTOR_SIZE  (CPQCILINESIZE)
#define CPQCI_PAGE_SIZE            (4*1024)  // 4k bytes
#define CPQCI_DOORBELL_SPAN_SHIFT  14        // log2 of 16k
#define CPQCI_DOORBELL_SPAN        (1 << CPQCI_DOORBELL_SPAN_SHIFT) // 16k bytes
#define CPQCI_SAWBASE_VIEWSIZE     0x800000  // should extract from device

/******************************************************************************
  The following offsets are into the PCI Function-2 configuration buffer.
  Note that the configuration buffer can also be accessed through IO and MMIO 
  cycles, relative to the BAR information found at offsets 0x10 and 0x14.
******************************************************************************/

#define CPQCI_REG_SIGNATURE    0x00    // 32b
#define CPQCI_REG_IOIOREGBAR   0x10    // 32b
#define CPQCI_REG_SHAREDRAMBAR 0x18    // 32b
#define CPQCI_REG_INTLINE      0x3C    //  8b
#define CPQCI_REG_INTPIN       0x3D    //  8b
#define CPQCI_REG_SHAREDRAMCFG 0x50    // 32b
#define CPQCI_REG_IRQENABLE    0xB2    // 16b
#define CPQCI_REG_DBINBOUND    0xD0    // 32b
#define CPQCI_REG_DBOUTBOUND   0xD4    // 32b

#ifndef CPQCI_DB_BITMAPPD
#define CPQCI_REG_DBPAGEBAR    0x1C
#else
#define CPQCI_REG_DBPAGEBAR    CPQCI_REG_MMIOREGBAR
#endif

/******************************************************************************
  The Channel Control Block (CCB) is used by the channel hardware to track
  channel state and contain information of where to send/receive messages.
  A CCB consists of 2 fifos and a context area used by HW & SW in any manner
  required to further support host messaging through the channel.

  The base addresses defined by SqFifo, SqDesc, RqFifo, RqDesc, and Doorbell 
  are valid relative to the caller's context.


  offset             CCB LAYOUT
          +-------------------------------+
  0x0000  |            SqFifo             |
          +-------------------------------+
  0x0008  |            SqDesc             |
          +-------------------------------+
  0x0010  |            SqCtrl             |
          +-------------------------------+
  0x0018  |            RqFifo             |
          +-------------------------------+
  0x0020  |            RqDesc             |
          +-------------------------------+
  0x0028  |            RqCtrl             |
          +-------------------------------+
  0x0030  |         DoorbellBase          |
          +-------------------------------+
  0x0038  |         ChannelNumber         |
          +-------------------------------+
  0x0040  |                               |
          |                               |
  0x0048  |                               |
          |                               |
  0x0050  |                               |
          |                               |
  0x0058  |                               |
          |            Context            |
  0x0060  |                               |
          |                               |
  0x0068  |                               |
          |                               |
  0x0070  |                               |
          |                               |
  0x0078  |                               |
          +-------------------------------+

******************************************************************************/

//  Define the offset from the device shared memory region
//  containing the base of the adapter's ccb table.
#define CPQCI_CCBTABLE_BASE  0x0000

typedef struct _CPQCICCB {
    //  send
    CPQCITYPEQUAD SqFifo;
    CPQCITYPEQUAD SqDesc;
    CPQCITYPEQUAD SqCtrl;
    //  recv
    CPQCITYPEQUAD RqFifo;
    CPQCITYPEQUAD RqDesc;
    CPQCITYPEQUAD RqCtrl;
    //  doorbell
    CPQCITYPEQUAD DoorbellBase;
    //  info
    CPQCITYPEQUAD ChannelNumber;
    //  context fields
    u8_t           Context[CPQCI_KEY_SIZE - (sizeof(CPQCITYPEQUAD) * 8)];
} CPQCICCB, CPQCIKEY;


/******************************************************************************
    Item to note...Although the MS compiler seems to generate fairly efficient
    code when accessing bit structures, other compilers seem to be void in the
    area of bit structure access optimizations.  Therefore, rather than 
    implementing bit structures and having a compiler (i.e. Greenhills) 
    generate a bazillion machine ops to do simple bit wise stuff, we will use
    masking and shifting techniques explicitly.
******************************************************************************/

//  Bit positions for fields within XxCtrl
#define CPQCI_CTRL_BITPOS_L2SZ             0
#define CPQCI_CTRL_BITPOS_FIFOINDEXMASK    4
#define CPQCI_CTRL_BITPOS_DESCLIMIT        18
#define CPQCI_CTRL_BITPOS_A                30
#define CPQCI_CTRL_BITPOS_G                31

//  Bits consumed by each field within XxCtrl
#define CPQCI_CTRL_BITS_L2SZ               4
#define CPQCI_CTRL_BITS_FIFOINDEXMASK      14
#define CPQCI_CTRL_BITS_DESCLIMIT          12
#define CPQCI_CTRL_BITS_A                  1
#define CPQCI_CTRL_BITS_G                  1
#define CPQCI_CTRL_BITS_TOTAL       \
   (CPQCI_CTRL_BITS_L2SZ           +\
    CPQCI_CTRL_BITS_FIFOINDEXMASK  +\
    CPQCI_CTRL_BITS_DESCLIMIT      +\
    CPQCI_CTRL_BITS_A              +\
    CPQCI_CTRL_BITS_G)

//  Bit field masks derived from above to provide field isolation.
#define CPQCI_CTRL_MASK_L2SZ\
    (((1 << CPQCI_CTRL_BITS_L2SZ) - 1) << CPQCI_CTRL_BITPOS_L2SZ)

#define CPQCI_CTRL_MASK_FIFOINDEXMASK\
    (((1 << CPQCI_CTRL_BITS_FIFOINDEXMASK) - 1) << CPQCI_CTRL_BITPOS_FIFOINDEXMASK)

#define CPQCI_CTRL_MASK_DESCLIMIT\
    (((1 << CPQCI_CTRL_BITS_DESCLIMIT) - 1) << CPQCI_CTRL_BITPOS_DESCLIMIT)

#define CPQCI_CTRL_MASK_A\
    (((1 << CPQCI_CTRL_BITS_A) - 1) << CPQCI_CTRL_BITPOS_A)

#define CPQCI_CTRL_MASK_G\
    (((1 << CPQCI_CTRL_BITS_G) - 1) << CPQCI_CTRL_BITPOS_G)

//  Macro to set all fields within the XxCtrl field.
#define CpqCiCtrlSetAll( _c, _l, _f, _d, _a, _g )         \
    *((vu_t *)(&(_c))) =                                   \
       ((((u_t)(_l)) << CPQCI_CTRL_BITPOS_L2SZ)          |\
        (((u_t)(_f)) << CPQCI_CTRL_BITPOS_FIFOINDEXMASK) |\
        (((u_t)(_d)) << CPQCI_CTRL_BITPOS_DESCLIMIT)     |\
        (((u_t)(_a)) << CPQCI_CTRL_BITPOS_A)             |\
        (((u_t)(_g)) << CPQCI_CTRL_BITPOS_G))

//  Macros to affect the 'G' bit within the XxCtrl field.
//  Note that references MUST be volatile to handle access
//  by 2 different asynchronous domains (i.e. hw & sw).
#define CpqCiCtrlSetGo( _ctrl )\
    *((vu_t *)(&(_ctrl))) |=  (1 << CPQCI_CTRL_BITPOS_G)
#define CpqCiCtrlClearGo( _ctrl )\
    *((vu_t *)(&(_ctrl))) &= ~(1 << CPQCI_CTRL_BITPOS_G)
#define CpqCiCtrlTestGo( _ctrl )\
   (*((vu_t *)(&(_ctrl))) & CPQCI_CTRL_MASK_G)

//  Macros to affect the 'A' bit within the XxCtrl field.
//  Note that references MUST be volatile to handle access
//  by 2 different asynchronous domains (i.e. hw & sw).
#define CpqCiCtrlSetActive( _ctrl )\
    *((vu_t *)(&(_ctrl))) |=  (1 << CPQCI_CTRL_BITPOS_A)
#define CpqCiCtrlClearActive( _ctrl )\
    *((vu_t *)(&(_ctrl))) &= ~(1 << CPQCI_CTRL_BITPOS_A)
#define CpqCiCtrlTestActive( _ctrl )\
   (*((vu_t *)(&(_ctrl))) & CPQCI_CTRL_MASK_A)

//  Macros to isolate and extract various XxCtrl fields.
#define CpqCiCtrlGetFifoindexmask( _ctrl )\
    ((*((vu_t *)(&(_ctrl))) & CPQCI_CTRL_MASK_FIFOINDEXMASK) >> CPQCI_CTRL_BITPOS_FIFOINDEXMASK)
#define CpqCiCtrlGetDesclimit( _ctrl )\
    ((*((vu_t *)(&(_ctrl))) & CPQCI_CTRL_MASK_DESCLIMIT) >> CPQCI_CTRL_BITPOS_DESCLIMIT)
#define CpqCiCtrlGetL2sz( _ctrl )\
    ((*((vu_t *)(&(_ctrl))) & CPQCI_CTRL_MASK_L2SZ) >> CPQCI_CTRL_BITPOS_L2SZ)
    
//  Macros to extract more relevant information from XxCtrl.
#define CpqCiDescMemorySize( _ctrl, _descs )\
    (((u_t)(_descs)) << CpqCiCtrlGetL2sz(_ctrl))

/******************************************************************************
  The following macros are used to perform copy and fill functions independent
  of OS dependent library routines.
******************************************************************************/

//  CpqCiCopyBuffer( DstAddress, SrcAddress, BytesToTransfer );
#define CpqCiCopyBuffer( _Dst, _Src, _Bytes )\
{\
    u8_t *_dst_ = ((void *)(_Dst));\
    u8_t *_src_ = ((void *)(_Src));\
    u_t    _len_ = ((u_t)(_Bytes));\
    while(_len_--) *_dst_++ = *_src_++;\
}

//  CpqCiFillBuffer( DstAddress, FillByte, BytesToFill );
#define CpqCiFillBuffer( _Dst, _Fil, _Bytes )\
{\
    u8_t *_dst_ = ((void *)(_Dst));\
    u_t    _len_ = ((u_t)(_Bytes));\
    while(_len_--) *_dst_++ = ((u8_t)(_Fil));\
}

//  CpqCiLog2( inNUM, outNUM, outEXP );
//      CpqCiLog2( 120, outNUM, outEXP );
//          outNUM result = 128
//          outEXP result =   7
#define CpqCiLog2( _I, _O, _E ) {(_O)=1;(_E)=0;while((_O)<(_I)){(_O)<<=1;(_E)++;}}

//  CpqCiAlignAddress( AddressToAlign, Alignment );
//      AlignedAddress = CpqCiAlignAddress( Address, LINE_SIZE );
#define CpqCiAlignAddress( _adr, _lin )\
    ( ((_lin) && (((u_t)(_adr)) & (((u_t)(_lin))-1)))                               ?\
        (void *)(&((char *)(_adr))[((u_t)(_lin))-(((u_t)(_adr))&(((u_t)(_lin))-1))]):\
        (void *)(_adr))

/******************************************************************************
 QWORDS     - Indicates the total number of qwords to consume from the 
              descriptor.
 DESCRIPTOR - The index number representing the descriptor within the specified
              Send/Recv descriptor pool region.
 'C'ONSUMED - Indicates that the 64-bit fifo entry has been used by the target.
              Target is just another name for the consumer or dequeuer.
 'O'CCUPIED - Indicates that the 64-bit fifo entry is occupied.
******************************************************************************/

#define CPQCI_ENTRY_BITPOS_QWORDS      0
#define CPQCI_ENTRY_BITPOS_DESCRIPTOR  10
#define CPQCI_ENTRY_BITPOS_C           22
#define CPQCI_ENTRY_BITPOS_O           23

#define CPQCI_ENTRY_BITS_C             1
#define CPQCI_ENTRY_BITS_O             1
#define CPQCI_ENTRY_BITS_QWORDS        10
#define CPQCI_ENTRY_BITS_DESCRIPTOR    12
#define CPQCI_ENTRY_BITS_TOTAL   \
   (CPQCI_ENTRY_BITS_C          +\
    CPQCI_ENTRY_BITS_O          +\
    CPQCI_ENTRY_BITS_QWORDS     +\
    CPQCI_ENTRY_BITS_DESCRIPTOR)

#define CPQCI_ENTRY_MASK\
    ((1 << CPQCI_ENTRY_BITS_TOTAL) - 1)
#define CPQCI_ENTRY_MASK_C\
    (((1 << CPQCI_ENTRY_BITS_C) - 1) << CPQCI_ENTRY_BITPOS_C)
#define CPQCI_ENTRY_MASK_O\
    (((1 << CPQCI_ENTRY_BITS_O) - 1) << CPQCI_ENTRY_BITPOS_O)
#define CPQCI_ENTRY_MASK_QWORDS\
    (((1 << CPQCI_ENTRY_BITS_QWORDS) - 1) << CPQCI_ENTRY_BITPOS_QWORDS)
#define CPQCI_ENTRY_MASK_DESCRIPTOR\
    (((1 << CPQCI_ENTRY_BITS_DESCRIPTOR) - 1) << CPQCI_ENTRY_BITPOS_DESCRIPTOR)

#define CpqCiEntrySetAll( _e, _d, _q )                  \
    *((vu_t *)(&(_e))) =                                 \
       ((((u_t)(_d)) << CPQCI_ENTRY_BITPOS_DESCRIPTOR) |\
        (((u_t)(_q)) << CPQCI_ENTRY_BITPOS_QWORDS))

#define CpqCiEntryGetQwords( _e )\
    ((((u_t)(_e)) & CPQCI_ENTRY_MASK_QWORDS) >> CPQCI_ENTRY_BITPOS_QWORDS)
#define CpqCiEntryGetDescriptor( _e )\
    ((((u_t)(_e)) & CPQCI_ENTRY_MASK_DESCRIPTOR) >> CPQCI_ENTRY_BITPOS_DESCRIPTOR)

#define CpqCiMakePacketAddress( _descbar, _ctrl, _desc )\
    ((CPQCIPACKET)(&(((CPQCITYPEQUAD *)(&(_descbar)))->ptr[CpqCiDescMemorySize(_ctrl,_desc)])))    

#define CpqCiChannelNumberGet( _ccb )\
    *((CPQCICHANNEL *)(&((CPQCICCB *)(_ccb))->ChannelNumber)) 
#define CpqCiChannelNumberSet( _ccb, _channel )\
    *((CPQCICHANNEL *)(&((CPQCICCB *)(_ccb))->ChannelNumber)) = ((CPQCICHANNEL)(_channel))

#define CpqCiAccessDoorbellAsInt( _ccb ) ((CPQCICCB *)(_ccb))->DoorbellBase.uint
#define CpqCiAccessDoorbellAsPtr( _ccb ) ((CPQCICCB *)(_ccb))->DoorbellBase.ptr

/******************************************************************************
    Simple LOOP construct for clarity.
******************************************************************************/

#define CpqCiEnterLoop for(;;){
#define CpqCiLeaveLoop }

/******************************************************************************
    This template is used to track and manage access to/from the
    fifo array.  If hardware parties on the fifo, then it will have its own 
    data structure to track and manage its access to the fifo array.
******************************************************************************/

typedef struct _CPQCIFIFO {
    // begin line
    CPQCITYPEQUAD rents;  // user requested number of fifo entries
    CPQCITYPEQUAD imask;  // mask to extract valid fifo index
    CPQCITYPEQUAD merge;  // O/C bits to merge in during enqueue operation
    CPQCITYPEQUAD reset;  // set to non-zero when the target device resets
    u8_t          pad_0[CPQCILINESIZE - (sizeof(CPQCITYPEQUAD) * 4)];
    // begin line
    vu64_t        head;
    u8_t          pad_1[CPQCILINESIZE - (sizeof(CPQCITYPEQUAD) * 1)];
    // begin line
    vu64_t        tail;
    u8_t          pad_2[CPQCILINESIZE - (sizeof(CPQCITYPEQUAD) * 1)];
    // begin line
    vu64_t        fifo[1];
} CPQCIFIFO, * CPQCIFIFOHANDLE;

#define CPQCIFIFOHANDLESIZE   (sizeof(CPQCIFIFO) - sizeof(vu64_t))

#define CpqCiFifoBar2FifoHandle( _fifo )\
    ((CPQCIFIFOHANDLE)(((u8_t *)(_fifo)) - CPQCIFIFOHANDLESIZE)) 

#define CpqCiAdapterReset( _registers )\
    (*((vu32_t *)(&((u8_t *)(_registers))[CPQCI_REG_DBOUTBOUND])) & (1 << 26))

#define CpqCiChannelReset( _ccb )\
   (CpqCiFifoBar2FifoHandle((_ccb)->SqFifo.ptr)->reset.uint ||\
    CpqCiFifoBar2FifoHandle((_ccb)->RqFifo.ptr)->reset.uint)

/******************************************************************************
  Prototypes for routines found in CpqCi.c.
  These routines can be called from user and kernel mode.
******************************************************************************/

CPQCISIZE
CpqCiFifoSize(
    CPQCIENTRIES FifoEntries,
    CPQCIENTRIES *pAdjustedEntries 
    );

CPQCIFIFOBAR 
CpqCiFifoSetup(
    CPQCIBUFFER HandleMemory,
    CPQCIENTRIES FifoEntries,
    CPQCIBOOL Normal
    );

CPQCIBOOL
CpqCiFifoEnqueue(
    CPQCIFIFOBAR Fifo,
    CPQCIENTRY Entry
    );

CPQCIBOOL
CpqCiFifoDequeue(
    CPQCIFIFOBAR Fifo,
    CPQCIENTRY *pEntry
    );

CPQCIBOOL
CpqCiFifoFull(
    CPQCIFIFOBAR Fifo
    );

CPQCIBOOL
CpqCiFifoEmpty(
    CPQCIFIFOBAR Fifo
    );

CPQCISTATUS
CpqCiCcbSetup(
    CPQCICCB     *Ccb,
    CPQCICHANNEL  ChannelNumber,
    CPQCIDOORBELL DoorbellBase,
    CPQCIFIFOBAR  SqFifoBar,
    CPQCIDESCBAR  SqDescBar,
    CPQCIENTRIES  SqDescs,
    CPQCISIZE     SqDescSize,
    CPQCICONTROL  SqGo,
    CPQCIFIFOBAR  RqFifoBar,
    CPQCIDESCBAR  RqDescBar,
    CPQCIENTRIES  RqDescs,
    CPQCISIZE     RqDescSize,
    CPQCICONTROL  RqGo
    );

CPQCIBOOL
CpqCiPacketEnqueue(
    CPQCICCB *Ccb,
    CPQCIFIFOTYPE TargetFifo,
    CPQCIPACKETID PacketId,
    CPQCILENGTH Length
    );

CPQCIBOOL
CpqCiPacketDequeue(
    CPQCICCB *Ccb,
    CPQCIFIFOTYPE TargetFifo,
    CPQCIPACKETID *pPacketId,
    CPQCILENGTH *pLength,
    CPQCIPACKET *pPacket
    );

CPQCIBUFFER
CpqCiStatusMessage(
    CPQCISTATUS StatusCode
    );

/******************************************************************************
  The following macros should not be used by the IOP software since these 
  macros assume a view from the Host interface.  Note that the CPQCI_SW_EMULATOR
  referred to here is the software emulator implemented on a non-os processor 
  "allocated" away from the host os.  If this code is to be implemented using 
  the hard channel emulator, then the CPQCI_SW_EMULATOR term needs to be 
  undefined (commented out) at the top of this file.
******************************************************************************/

#ifdef  CPQCI_DB_BITMAPPD
#define CpqCiGetDoorbellOffset( _ccb ) (CPQCI_REG_DBINBOUND)
#else
#define CpqCiGetDoorbellOffset( _ccb ) (CpqCiChannelNumberGet(_ccb)<<CPQCI_DOORBELL_SPAN_SHIFT)
#endif

#ifdef  CPQCI_SW_EMULATOR
#define CpqCiDoorbellSet( _ccb )\
{\
    u_t __c =  (1 << CpqCiChannelNumberGet( _ccb ));\
    void * __db = (void *)CpqCiAccessDoorbellAsPtr( _ccb );\
    _asm{      mov  edx, dword ptr __db     }\
    _asm{      mov  eax, dword ptr __c      }\
    _asm{ lock or   dword ptr [edx], eax    }\
}
#define CpqCiDoorbellClr( _ccb )\
{\
    u_t __c = ~(1 << CpqCiChannelNumberGet( _ccb ));\
    void * __db = (void *)CpqCiAccessDoorbellAsPtr( _ccb );\
    _asm{      mov  edx, dword ptr __db     }\
    _asm{      mov  eax, dword ptr __c      }\
    _asm{ lock and  dword ptr [edx+4], eax  }\
}
#else   // REAL THING HERE
#ifdef  CPQCI_DB_BITMAPPD
#define CpqCiDoorbellSet( _ccb )\
    ((vu32_t *)(CpqCiAccessDoorbellAsPtr(_ccb)))[0]=(1<<CpqCiChannelNumberGet(_ccb))
#define CpqCiDoorbellClr( _ccb )\
    ((vu32_t *)(CpqCiAccessDoorbellAsPtr(_ccb)))[4]=(1<<CpqCiChannelNumberGet(_ccb))
#else
#define CpqCiDoorbellSet( _ccb ) *((vu32_t *)(CpqCiAccessDoorbellAsPtr(_ccb))) = 0x1
#define CpqCiDoorbellClr( _ccb ) *((vu32_t *)(CpqCiAccessDoorbellAsPtr(_ccb))) = 0x2
#endif
#endif

#define CpqCiDoorbellMask( _ccb )   ((vu8_t *)(&((CPQCICCB *)(_ccb))->RqCtrl))[4] = 0
#define CpqCiDoorbellUnmask( _ccb ) ((vu8_t *)(&((CPQCICCB *)(_ccb))->RqCtrl))[4] = 1

/******************************************************************************
  Prototypes for routines found in CpqCiUxx.c & CpqCiKxx.c,
  where "xx" indicates the target host operating system.
******************************************************************************/

CPQCISTATUS
CpqCiInitialize(
    CPQCICONTEXT Context
    );

CPQCISTATUS
CpqCiCreate(
    CPQCIKEY *pKey,
    CPQCIHANDLE *pHandle,
    CPQCIENTRIES SendPackets,
    CPQCISIZE SendPacketSize,
    CPQCIENTRIES RecvPackets,
    CPQCISIZE RecvPacketSize,
    CPQCIPRIORITY Priority,
    CPQCITIMEOUT *pWaitTime
    );
    
CPQCISTATUS
CpqCiOpen(
    CPQCIKEY *pKey,
    CPQCIHANDLE *pHandle
    );
    
CPQCISTATUS
CpqCiClose(
    CPQCIHANDLE Handle
    );

CPQCILENGTH
CpqCiSend(
    CPQCIHANDLE Handle,
    CPQCIBUFFER Buffer,
    CPQCILENGTH Length,
    CPQCISTATUS *pError,
    CPQCITIMEOUT *pWaitTime
    );

CPQCILENGTH
CpqCiRecv(
    CPQCIHANDLE Handle,
    CPQCIBUFFER Buffer,
    CPQCILENGTH Length,
    CPQCISTATUS *pError,
    CPQCITIMEOUT *pWaitTime
    );


void DumpCpqCiCcb(CPQCICCB* x);
void DumpCpqCiFifoHandle(CPQCIFIFOHANDLE x);
				     
#endif // _CPQCIh
