/*++

(C) Copyright 2000-2006 Hewlett-Packard Development Company, L.P.

Module Name:

    CpqCiDrv.c

Abstract:

    This module contains code to manipulate channels from OS-dependent 
    device drivers.

Environment:

    OS Device driver context.

Revision History:

     11/07/2006  PMS   First Open Source Release


--*/

#include "CpqCiWrp.h"
#include "CpqCiTyp.h"
#include "CpqCiDrv.h"
#include "CpqCi.h"

/*****************************************************************************
    Global version string.
******************************************************************************/

const char *iLoVersion="ILO_VERSION=1.00 - 09.04.2002";// 09/04/2002 tjb: (b)


/*****************************************************************************
    Prototypes for private support routines used solely by this module.
******************************************************************************/

CPQCIPRIVATE
CPQCISTATUS
MapConnection( 
    CPQCICONTEXT Context,
    CPQCIADAPTER Adapter,
    CPQCICHANNEL Channel,
    CPQCIMODE AddressMode,
    CPQCICCB *pCcb,
    CPQCIHANDLE *pMemoryHandle,
    CPQCIHANDLE *pDoorbellHandle
    );

CPQCIPRIVATE
CPQCIVOID
ConstructKey(
    CPQCIBUFFER pDst,
    CPQCIBUFFER pSrc
    );

/*****************************************************************************
    Exported routines.
******************************************************************************/

CPQCISTATUS
CpqCiDrvCreateAdapter(
    CPQCICONTEXT Context,
    CPQCIPHYSICAL RegisterBarPa,
    CPQCISIZE RegisterBarSize,
    CPQCICONNECTIONS MaxConnections,
    CPQCIADAPTER *pAdapter
    )

/*++

Routine Description:

    This routine creates an adapter object for the target device.

Arguments:

    Context - The caller's supplied context information required for some OS
        dependent functions such as the CpqCiDrvOsXx() memory management 
        routines.

    RegisterBarPa - The physical address of the target device's memory mapped
        I/O registers.

    RegisterBarSize - The size in bytes of the memory mapped I/O register 
        region.

    MaxConnections - The total number of connections for the target device.
        This value limits the total number of PROCESS connections that can
        perform I/O to the target device.  This value does not limit the 
        total number of threads within a PROCESS that can perform I/O.

    pAdapter - Location to place the adapter object that describes the
        target device.
    
Return Value:

    The result of the operation.  

Caveats:

    *pAdapter is only valid when "CpqCiSuccess" is returned.

--*/
{
    CPQCICHANNELS maxconnections = MaxConnections;
    CPQCICHANNELS maxchannels = CPQCI_MAX_CHANNELS; //bugbug: get this from hw.
    CPQCIADAPTEROBJECT *pao = CPQCINULL;
    CPQCIFIFOBAR qpriority = CPQCINULL;
    CPQCISTATUS error = CpqCiSuccess;
    CPQCIHANDLE handle;
    CPQCIPHYSICAL pa;
    CPQCIENTRY entry;
    CPQCISIZE size;
    CPQCIBYTE *va;

    /*************************************************************************
        Compute the total size of memory required for the adapter object.
    **************************************************************************/

    //  Compute static template sizes.
    size  = (sizeof(CPQCIADAPTEROBJECT) * 1) + CPQCILINESIZE;
    size += (sizeof(CPQCICHANNELOBJECT) * maxchannels) + CPQCILINESIZE;
    size += (sizeof(CPQCICONNECTIONOBJECT) * maxconnections) + CPQCILINESIZE;

    //  Include fifo Q sizes.  Computations already include linesize alignment.
    size += CpqCiFifoSize(maxconnections, CPQCINULL);
    size += CpqCiFifoSize(CPQCI_NORMPRI_CHANNELS, CPQCINULL);
    size += CpqCiFifoSize(CPQCI_HIGHPRI_CHANNELS, CPQCINULL);
    size += CpqCiFifoSize(CPQCI_CRI0PRI_CHANNELS, CPQCINULL);
    size += CpqCiFifoSize(CPQCI_CRI1PRI_CHANNELS, CPQCINULL);
    size += CpqCiFifoSize(CPQCI_CRI2PRI_CHANNELS, CPQCINULL);
    size += CpqCiFifoSize(CPQCI_CRI3PRI_CHANNELS, CPQCINULL);
    size += CpqCiFifoSize(CPQCI_CRI4PRI_CHANNELS, CPQCINULL);
    size += CpqCiFifoSize(CPQCI_CRI5PRI_CHANNELS, CPQCINULL);
    size += CpqCiFifoSize(CPQCI_CRI6PRI_CHANNELS, CPQCINULL);
    size += CpqCiFifoSize(CPQCI_CRI7PRI_CHANNELS, CPQCINULL);

    /*************************************************************************
        Allocate the adapter object memory region.
    **************************************************************************/

    DBG("AdapterObject size = %d (hex %p)\n", size, size);
    handle = CpqCiDrvOsMalloc(Context, size);
    if( !handle )
    {
        error = CpqCiErrNoMemory;
        goto exitProc;
    }

    /*************************************************************************
        Extract the virtual address for the adapter object and begin
        dividing it into sections.
    **************************************************************************/

    va = CpqCiDrvOsGetVA(Context, handle);

    //  The begining of this region will be where the adapter object fields
    //  reside.  However, ensure this address begins on a linesize alignment.
    pao = CpqCiAlignAddress(va, CPQCILINESIZE);

    //  Update what we can of the adapter object.
    pao->MaxConnections = maxconnections;
    pao->MaxChannels = maxchannels;
    pao->AdapterObjMemHandle = handle;

    //  Channel object region is after the adapter object memory.
    va = &va[(sizeof(CPQCIADAPTEROBJECT) * 1) + CPQCILINESIZE];
    pao->ChannelObject = CpqCiAlignAddress(va, CPQCILINESIZE);
    CpqCiMemSet(pao->ChannelObject, 0, sizeof(CPQCIADAPTEROBJECT));

    //  Connection object region is after the channel object memory.
    va = &va[(sizeof(CPQCICHANNELOBJECT) * maxchannels) + CPQCILINESIZE];
    pao->ConnectionObject = CpqCiAlignAddress(va, CPQCILINESIZE);
    CpqCiMemSet(pao->ChannelObject, 0, sizeof(CPQCICHANNELOBJECT) * maxchannels);

    //  Fifo Q's follow everything else.
    va = &va[(sizeof(CPQCICONNECTIONOBJECT) * maxconnections) + CPQCILINESIZE];
    pao->QConnectionObjs = CpqCiFifoSetup(va, maxconnections, CpqCiTrue);
    CpqCiMemSet(pao->ChannelObject, 0, sizeof(CPQCICONNECTIONOBJECT) * maxconnections);

    va = &va[CpqCiFifoSize(maxconnections, CPQCINULL)];
    pao->QNormPriority = CpqCiFifoSetup(va, CPQCI_NORMPRI_CHANNELS, CpqCiTrue);

    va = &va[CpqCiFifoSize(CPQCI_NORMPRI_CHANNELS, CPQCINULL)];
    pao->QHighPriority = CpqCiFifoSetup(va, CPQCI_HIGHPRI_CHANNELS, CpqCiTrue);

    va = &va[CpqCiFifoSize(CPQCI_HIGHPRI_CHANNELS, CPQCINULL)];
    pao->QCri0Priority = CpqCiFifoSetup(va, CPQCI_CRI0PRI_CHANNELS, CpqCiTrue);

    va = &va[CpqCiFifoSize(CPQCI_CRI0PRI_CHANNELS, CPQCINULL)];
    pao->QCri1Priority = CpqCiFifoSetup(va, CPQCI_CRI1PRI_CHANNELS, CpqCiTrue);

    va = &va[CpqCiFifoSize(CPQCI_CRI1PRI_CHANNELS, CPQCINULL)];
    pao->QCri2Priority = CpqCiFifoSetup(va, CPQCI_CRI2PRI_CHANNELS, CpqCiTrue);

    va = &va[CpqCiFifoSize(CPQCI_CRI2PRI_CHANNELS, CPQCINULL)];
    pao->QCri3Priority = CpqCiFifoSetup(va, CPQCI_CRI3PRI_CHANNELS, CpqCiTrue);

    va = &va[CpqCiFifoSize(CPQCI_CRI3PRI_CHANNELS, CPQCINULL)];
    pao->QCri4Priority = CpqCiFifoSetup(va, CPQCI_CRI4PRI_CHANNELS, CpqCiTrue);

    va = &va[CpqCiFifoSize(CPQCI_CRI4PRI_CHANNELS, CPQCINULL)];
    pao->QCri5Priority = CpqCiFifoSetup(va, CPQCI_CRI5PRI_CHANNELS, CpqCiTrue);

    va = &va[CpqCiFifoSize(CPQCI_CRI5PRI_CHANNELS, CPQCINULL)];
    pao->QCri6Priority = CpqCiFifoSetup(va, CPQCI_CRI6PRI_CHANNELS, CpqCiTrue);

    va = &va[CpqCiFifoSize(CPQCI_CRI6PRI_CHANNELS, CPQCINULL)];
    pao->QCri7Priority = CpqCiFifoSetup(va, CPQCI_CRI7PRI_CHANNELS, CpqCiTrue);

    /*************************************************************************
        Map the adapter hardware register block into kernel space.
    **************************************************************************/

    pao->MmioRegistersHandle =
    CpqCiDrvOsMap( Context,
                   RegisterBarPa & ~0xf,
                   RegisterBarSize,
                   CpqCiKernelMode,
                   CpqCiUncached );

    if( !pao->MmioRegistersHandle )
    {
        error = CpqCiErrMapRegisters;
        goto exitProc;
    }

    va = CpqCiDrvOsGetVA(Context, pao->MmioRegistersHandle);

    //  Make sure the caller is pointing us to a valid adapter.
    if( *((vu32_t *)(&va[CPQCI_REG_SIGNATURE])) != CPQCI_SIGNATURE )
    {
        error = CpqCiErrInvalidAdapter;
        goto exitProc;
    }

    //  Squelch any device chatter via interrupts.
    *((vu32_t *)(&va[CPQCI_REG_DBOUTBOUND])) = ((u32_t)(-1));

    /*************************************************************************
        Map the adapter hardware doorbell page aperature region into 
        kernel space.
    **************************************************************************/

    pa = (CPQCIPHYSICAL)(*((vu32_t *)(&va[CPQCI_REG_DBPAGEBAR])) & ~0xf);
    size = (CPQCISIZE)(maxchannels * CPQCI_DOORBELL_SPAN);

    pao->MmioDoorbellHandle =
    CpqCiDrvOsMap( Context,
                   pa,
                   size,
                   CpqCiKernelMode,
                   CpqCiUncached );

    if( !pao->MmioDoorbellHandle )
    {
        error = CpqCiErrMapDoorbells;
        goto exitProc;
    }

    /*************************************************************************
        Map the adapter hardware shared memory region into kernel space.
        We only need to map the region containing the CCB table.
    **************************************************************************/

    pa = (CPQCIPHYSICAL)(*((vu32_t *)(&va[CPQCI_REG_SHAREDRAMBAR])) & ~0xf);
    size = (CPQCISIZE)(*((vu32_t *)(&va[CPQCI_REG_SHAREDRAMCFG])) & 0xf);

    //  Adjust the reported size if not zero.  The device must have some ram.
    if( !size )
    {
        error = CpqCiErrNoDeviceMemory;
        goto exitProc;
    }

    size = 0x1000 << (size - 1);

    //  We only need to map the ccb table.
    pao->MmioCcbTableHandle = 
    CpqCiDrvOsMap( Context,
                   pa + CPQCI_CCBTABLE_BASE,
                   sizeof(CPQCICCB) * maxchannels,
                   CpqCiKernelMode,
                   CpqCiUncached );

    DBG("MmioCcbTableHandle = %x\n", pao->MmioCcbTableHandle ); 

    if( !pao->MmioCcbTableHandle )
    {
        error = CpqCiErrMapDeviceram;
        goto exitProc;
    }

    /*************************************************************************
        Add the connection entries to the connection object Q.
    **************************************************************************/

    for( entry = 0; entry < maxconnections; entry++ )
        CpqCiFifoEnqueue(pao->QConnectionObjs, entry);
        
    /*************************************************************************
        Add channels to their appropriate priority queues.
    **************************************************************************/

    for( entry = 0; entry < maxchannels; entry++ )
    {
        //  Assign the correct priority queue.
        if( entry < CPQCI_BASE_HIGHPRIORITY_CHANNEL )
            qpriority = pao->QNormPriority;
        else
        if( entry < CPQCI_BASE_CRITICALPRIORITY_CHANNEL )
            qpriority = pao->QHighPriority;
        else
        switch( entry )
        {
            case CpqCiCriticalPriority_0:
                qpriority = pao->QCri0Priority;
                break;
            case CpqCiCriticalPriority_1:
                qpriority = pao->QCri1Priority;
                break;
            case CpqCiCriticalPriority_2:
                qpriority = pao->QCri2Priority;
                break;
            case CpqCiCriticalPriority_3:
                qpriority = pao->QCri3Priority;
                break;
            case CpqCiCriticalPriority_4:
                qpriority = pao->QCri4Priority;
                break;
            case CpqCiCriticalPriority_5:
                qpriority = pao->QCri5Priority;
                break;
            case CpqCiCriticalPriority_6:
                qpriority = pao->QCri6Priority;
                break;
            case CpqCiCriticalPriority_7:
                qpriority = pao->QCri7Priority;
                break;
            default:
                qpriority = CPQCINULL;

                break;
        }

        //  Place the channel number on the appropriate channel fifo Q.
        if( qpriority )
            CpqCiFifoEnqueue(qpriority, entry);

    }//end for( entry = 0; entry < maxchannels; entry++ )

    /*************************************************************************
        Enable PCI INTB# (pin 2) doorbell interrupts.
    **************************************************************************/

    *((vu8_t *)(&va[CPQCI_REG_IRQENABLE])) |= 1;

    /*************************************************************************
        Leave the routine and finish appropriately.
    **************************************************************************/

    error = CpqCiSuccess;

exitProc:
    //  Pick up our toys if an error occurred.
    if( error != CpqCiSuccess )
    {
        CpqCiDrvDeleteAdapter(Context, pao);
        pao = CPQCINULL;
    }

    *pAdapter = pao;

    return error;
}

CPQCISTATUS
CpqCiDrvDeleteAdapter(
    CPQCICONTEXT Context,
    CPQCIADAPTER Adapter
    )

/*++

Routine Description:

    This routine deletes the specified adapter object.

Arguments:

    Context - The caller's supplied context information required for 
        some OS dependent functions such as the CpqCiDrvOsXx() memory 
        management routines.

    Adapter - A valid adapter object returned from a previous call to
        CpqCiDrvCreateAdapter().
    
Return Value:

    The result of the operation.

Caveats:

    None.

--*/
{
    if( Adapter )
    {
        // 09/10/2001 tjb: (b)
        u8_t *va = CpqCiDrvOsGetVA(Context, Adapter->MmioRegistersHandle);

        //  Squelch any device chatter via interrupts.  09/10/2001 tjb: (b)
        *((vu8_t  *)(&va[CPQCI_REG_IRQENABLE])) &= ((u8_t)(~1));
        *((vu32_t *)(&va[CPQCI_REG_DBOUTBOUND])) = ((u32_t)(-1));

        //  Free up any allocated/mapped resources.  The CpqCiDrvOsFree()
        //  routine will test for a NULL handle value.
        CpqCiDrvOsFree(Context, Adapter->MmioCcbTableHandle);
        CpqCiDrvOsFree(Context, Adapter->MmioDoorbellHandle);
        CpqCiDrvOsFree(Context, Adapter->MmioRegistersHandle);
        CpqCiDrvOsFree(Context, Adapter->AdapterObjMemHandle);
    }

    return CpqCiSuccess;
}

CPQCISTATUS
CpqCiDrvCreateChannel(
    CPQCICONTEXT Context,
    CPQCIADAPTER Adapter,
    CPQCIENTRIES SendPackets,
    CPQCISIZE SendPacketSize,
    CPQCIENTRIES RecvPackets,
    CPQCISIZE RecvPacketSize,
    CPQCICCB *CallersCcb,
    CPQCIPRIORITY ChannelPriority,
    CPQCIMODE AddressMode,
    CPQCIKEY *pKey,
    CPQCICONNECTION *pConnection,
    CPQCICHANNEL *pChannel
    )

/*++

Routine Description:

    This routine creates an exclusive connection to a channel based 
    on the specified priority.

Arguments:

    Context - The caller's supplied context information required 
        for some OS dependent functions such as the CpqCiDrvOsXx() 
        memory management routines.

    Adapter - Target adapter object.

    SendPackets - The number of send descriptors.
    SendPacketSize - The size in bytes for each send descriptor.

    RecvPackets - The number of recv descriptors.
    RecvPacketSize - The size in bytes for each recv descriptor.

    CallersCcb - Pointer to caller's CCB filled in with context 
        specific address information to operate on the channel 
        backing the connection.  This buffer will only get filled
        in if not NULL.

    ChannelPriority - The requested priority for the target channel.  
        Valid values are the following.

            CpqCiNormalPriority
            CpqCiHighPriority

        The following priorities are only valid from a kernel mode caller:
        
            CpqCiCriticalPriority_0
            CpqCiCriticalPriority_1
            CpqCiCriticalPriority_2
            CpqCiCriticalPriority_3
            CpqCiCriticalPriority_4
            CpqCiCriticalPriority_5
            CpqCiCriticalPriority_6
            CpqCiCriticalPriority_7

    AddressMode - The requested address mode to map the connection 
        address information.  Valid values for this field are the following.

            CpqCiUserMode
            CpqCiKernelMode

    pKey - Pointer to the caller's key buffer that will receive the key 
        created by this routine for this channel.  If pKey == NULL, then 
        the key will not be returned to the caller and the caller will 
        not beable to share this channel with other processes.

    pConnection - Location to place the connection created for this channel.

    pChannel - Location to place the 0-based channel number associated with 
        this connection.  If the value at this location is -1, the channel
        was never removed from the free list.  Otherwise, the channel was a
        valid channel number removed and then replaced on the free list
        if an error occurred.

Return Value:

    The result of the operation.
    
Caveats:

    If this routine does not return CpqCiSuccess, then the caller must
    check the *pChannel location to determine if a channel was actually
    removed from the free list, possibly causing other channel requestors
    of the same priority to block on an empty list.  For example, a typical
    caller of this routine attempting to implement channel signalling would
    implement the code as follows...

        CPQCISTATUS error;
        .
        .
        .
        error = CpqCiDrvCreateChannel( Context,
                                       Adapter,
                                       SendPackets,
                                       SendPacketSize,
                                       RecvPackets,
                                       RecvPacketSize,
                                       &CallersCcb,
                                       ChannelPriority,
                                       AddressMode,
                                       &Key,
                                       &Connection,
                                       &Channel );

        if( error != CpqCiSuccess )
        {
            if( Channel != CPQCI_INVALID_CHANNEL )
            {
                //  A channel was removed from the free list
                //  but had to be replaced as a result of an
                //  error.  Signal possible waiters that a 
                //  channel was placed back onto the free list
                //  for the target priority.
                SignalEvent(...);
            }
        }
        else
            ...

    The CpqCiCriticalPriority_0 through CpqCiCriticalPriority_7 values
    map to the critical priority channels 24 through 31 being requested
    for access.  These channels are special and will not export to 
    user-mode clients.

--*/
{
    CPQCIADAPTEROBJECT *pao = Adapter;
    CPQCICCB *ccbtbl = CpqCiDrvOsGetVA(Context, Adapter->MmioCcbTableHandle);
    CPQCICCB *driverccb = CPQCINULL;
    CPQCICONNECTION connection = CPQCI_INVALID_CONNECTION;
    CPQCICHANNEL channel = CPQCI_INVALID_CHANNEL;
    CPQCIFIFOBAR qpriority = CPQCINULL;
    CPQCISTATUS error = CpqCiSuccess;
    CPQCIHANDLE tmphandle = CPQCINULL;
    CPQCIHANDLE handle = CPQCINULL;
    CPQCIPACKETID packetid;
    CPQCISIZE packetsize;
    CPQCITYPEQUAD vabase;
    CPQCITYPEQUAD va;
    CPQCITYPEQUAD pabase;
    CPQCITYPEQUAD pa;
    CPQCITYPEQUAD size;
    u_t attempts;
    u8_t *pregs = CpqCiDrvOsGetVA(Context, pao->MmioRegistersHandle);

    DBG("\n CpqCiDrvCreateChannel ENTER:\n")
    DBG(" Adapter = pao = %p\n", pao);

    /*************************************************************************
        Make sure user-mode callers cannot access critical priority channels.
    **************************************************************************/

    if( AddressMode == CpqCiUserMode && ChannelPriority > CpqCiHighPriority )
    {
        error = CpqCiErrBadPriority;
        goto exitProc;
    }

    /*************************************************************************
        Determine which channel queue to obtain a free channel from.  This
        must be based on the requested priority.
    **************************************************************************/

    switch( ChannelPriority )
    {
        case CpqCiNormalPriority:
            qpriority = pao->QNormPriority;
            break;
        case CpqCiHighPriority:
            qpriority = pao->QHighPriority;
            break;
        case CpqCiCriticalPriority_0:
            qpriority = pao->QCri0Priority;
            break;
        case CpqCiCriticalPriority_1:
            qpriority = pao->QCri1Priority;
            break;
        case CpqCiCriticalPriority_2:
            qpriority = pao->QCri2Priority;
            break;
        case CpqCiCriticalPriority_3:
            qpriority = pao->QCri3Priority;
            break;
        case CpqCiCriticalPriority_4:
            qpriority = pao->QCri4Priority;
            break;
        case CpqCiCriticalPriority_5:
            qpriority = pao->QCri5Priority;
            break;
        case CpqCiCriticalPriority_6:
            qpriority = pao->QCri6Priority;
            break;
        case CpqCiCriticalPriority_7:
            qpriority = pao->QCri7Priority;
            break;
        default:
            qpriority = CPQCINULL;
            error = CpqCiErrBadPriority;
            goto exitProc;
    }

    /*************************************************************************
        Try to remove a channel from the target priority Q.  If none are 
        remaining, then indicate that all channels of this priority are
        in use.
    **************************************************************************/

    if( !CpqCiFifoDequeue(qpriority, (CPQCIENTRY *)&channel) )
    {
	DBG("CpqCiFifoDequeue failed to obtain a channel! - Critical Error!!\n");
        channel = CPQCI_INVALID_CHANNEL;
        error = CpqCiErrNoMoreChannels;
        goto exitProc;
    }

    pao->ChannelObject[channel].QPriority = qpriority;

    /*************************************************************************
        Try to obtain a connection object number for the requested channel.
    **************************************************************************/

    if( !CpqCiFifoDequeue(pao->QConnectionObjs, (CPQCIENTRY *)&connection) )
    {
	DBG("CpqCiFifoDequeue failed to obtain a connection! - Critical Error!!\n");
        connection = CPQCI_INVALID_CONNECTION;
        error = CpqCiErrNoMoreConnections;
        goto exitProc;
    }

    /*************************************************************************
        Validate the caller supplied ccb parameters.  Use the driver's ccb
        located within our channel object for this channel.
    **************************************************************************/

    driverccb = &pao->ChannelObject[channel].DriverCcb;

    error = 
    CpqCiCcbSetup( driverccb,
                   channel,
                   CPQCINULL,
                   CPQCINULL,CPQCINULL,SendPackets,SendPacketSize,0,
                   CPQCINULL,CPQCINULL,RecvPackets,RecvPacketSize,0 );

    if( error != CpqCiSuccess )
        goto exitProc;

    /*************************************************************************
        Compute the size of the physical contiguous host memory region used
        to back the channel.  We do not consider mapping the hardware shared
        ram region into the caller's address domain since it has been 
        discussed in various engineering meetings to consider the remapping
        of the adapter's physical resources exposed to the host.  The general
        layout of the host memory region follows...

                +-------------------------------------------+
                |                                           |
                |             SEND FIFO MEMORY              |
                |                                           |
                +-------------------------------------------+
                |                                           |
                |             RECV FIFO MEMORY              |
                |                                           |
                +-------------------------------------------+
                |                                           |
                |                                           |
                |                                           |
                |          SEND DESCRIPTOR MEMORY           |
                |                                           |
                |                                           |
                |                                           |
                +-------------------------------------------+
                |                                           |
                |                                           |
                |                                           |
                |          RECV DESCRIPTOR MEMORY           |
                |                                           |
                |                                           |
                |                                           |
                +-------------------------------------------+
                
    **************************************************************************/

    size.uint  = CpqCiFifoSize( SendPackets, CPQCINULL );
    size.uint += CpqCiFifoSize( RecvPackets, CPQCINULL );
    size.uint += CpqCiDescMemorySize( driverccb->SqCtrl, SendPackets );
    size.uint += CpqCiDescMemorySize( driverccb->RqCtrl, RecvPackets );

    //  Round out the size to consume pages and add in an extra page
    //  to account for any allocation we receive that does not map to
    //  a page boundary and requires us to align it.  The reason for all
    //  this madness is to keep the channel memory within an isolated
    //  fault zone in the case a rogue user-mode application attempts to
    //  party on the memory in a malicious manner.
    size.ptr   = CpqCiAlignAddress( size.ptr, CPQCI_PAGE_SIZE );
    size.uint += CPQCI_PAGE_SIZE;

    /*************************************************************************
        Allocate the host memory used to back the channel.  The allocation
        must take place in such a manner as to ensure that the memory region
        can be fully contained within the target adapter's SAWBASE register
        view.
    **************************************************************************/

    do
    {
        u_t sawbaseview = CPQCI_SAWBASE_VIEWSIZE; //bugbug: get from device.
        u_t startadr;
        u_t endadr;

        handle = CpqCiDrvOsMallocContiguous( Context, size.uint );
	DBG("CpqCiDrvOsMallocContiguous returned %p\n", handle);

        //  If there was a previous allocation, free it & reset tmphandle.
        CpqCiDrvOsFree( Context, tmphandle );
        tmphandle = CPQCINULL;

        if( !handle )
        {
	    DBG("CpqCiDrvOsMallocContiguous FAILED!!\n");
            error = CpqCiErrNoMoreChannelMemory;
            goto exitProc;
        }

        //  Extract the virtual address from the allocated memory handle and
        //  align it up to the next page.  This is where we will start dividing
        //  up the memory for the fifo queues and descriptor memory.  Account
        //  for the adjustment in size.
        vabase.ptr = CpqCiDrvOsGetVA( Context, handle );
        va.ptr = CpqCiAlignAddress( vabase.ptr, CPQCI_PAGE_SIZE );

        pabase.uint = CpqCiDrvOsGetPA( Context, handle );
        pa.ptr = CpqCiAlignAddress( pabase.ptr, CPQCI_PAGE_SIZE );

        size.uint -= (va.uint - vabase.uint);

        //  Extract the start and end address information and determine if
        //  the allocated memory buffer is within the sawbase view for the
        //  adapter.
//        startadr =  va.uint & ~(sawbaseview-1);                 // 09/04/2002 tjb: (a)
//        endadr   = (va.uint + (size.uint-1)) & ~(sawbaseview-1);// 09/04/2002 tjb: (a)
        startadr =  pa.uint & ~(sawbaseview-1);                 // 09/04/2002 tjb: (a)
        endadr   = (pa.uint + (size.uint-1)) & ~(sawbaseview-1);// 09/04/2002 tjb: (a)
            
        //  If the start and end address pages do not match, then the
        //  allocated buffer will not fit within a single sawbase register
        //  view for the adapter.  Hold onto the current allocated buffer 
        //  address and resume another allocation attempt to try and get a 
        //  buffer address that fulfills this requirement.
        if( startadr != endadr )
            tmphandle = handle;

    }while( tmphandle );

    //  Compute the doorbell page aperature address for this channel.
    DBG("Seeking VA for doorbell page\n");
    vabase.ptr  = CpqCiDrvOsGetVA( Context, pao->MmioDoorbellHandle );
    DBG("Seeking PA for doorbell page\n");
    pabase.uint = CpqCiDrvOsGetPA( Context, pao->MmioDoorbellHandle );

    vabase.ptr += CpqCiGetDoorbellOffset( driverccb );
    pabase.ptr += CpqCiGetDoorbellOffset( driverccb );

    //  Assign the channel object fields.
    pao->ChannelObject[channel].MemoryHandle = handle;
    pao->ChannelObject[channel].MemorySizeAdj = size.uint;
    pao->ChannelObject[channel].MemoryVaAdj.ptr = va.ptr;
    pao->ChannelObject[channel].MemoryPaAdj.ptr = pa.ptr;
    pao->ChannelObject[channel].DbBaseVaAdj.ptr = vabase.ptr;
    pao->ChannelObject[channel].DbBasePaAdj.ptr = pabase.ptr;

    /*************************************************************************
        Compute CCB offset information and initialize channel fifos.
    **************************************************************************/
    
    //  Compute the ccb's address information relative to the base memory 
    //  supplied by the caller.
    vabase.ptr = va.ptr;

    DBG("Calling CpqCiFifoSetup for SendPackets . . .\n");
    size.ptr = CpqCiFifoSetup( va.ptr, SendPackets, 0 );
    pao->ChannelObject[channel].SqFifoOffset.uint = size.uint - vabase.uint;
    va.ptr += CpqCiFifoSize(SendPackets, CPQCINULL);

    DBG("Calling CpqCiFifoSetup for RecvPackets . . .\n");
    size.ptr = CpqCiFifoSetup( va.ptr, RecvPackets, 0 );
    pao->ChannelObject[channel].RqFifoOffset.uint = size.uint - vabase.uint;
    va.ptr += CpqCiFifoSize(RecvPackets, CPQCINULL);

    size.ptr = va.ptr;
    pao->ChannelObject[channel].SqDescOffset.uint = size.uint - vabase.uint;
    va.ptr += CpqCiDescMemorySize(driverccb->SqCtrl, SendPackets);

    size.ptr = va.ptr;
    pao->ChannelObject[channel].RqDescOffset.uint = size.uint - vabase.uint;

    /*************************************************************************
        Setup the driver's ccb with virtual kernel mode addresses.
    **************************************************************************/

    DBG("Calling CpqCiCcbSetup for Virtual Kernel Mode Addresses . . .\n");
    CpqCiCcbSetup( driverccb,
                   channel,
                   pao->ChannelObject[channel].DbBaseVaAdj.ptr,
                   pao->ChannelObject[channel].MemoryVaAdj.ptr +
                        pao->ChannelObject[channel].SqFifoOffset.uint,
                   pao->ChannelObject[channel].MemoryVaAdj.ptr +
                        pao->ChannelObject[channel].SqDescOffset.uint,
                   SendPackets,
                   SendPacketSize,
                   1,
                   pao->ChannelObject[channel].MemoryVaAdj.ptr +
                        pao->ChannelObject[channel].RqFifoOffset.uint,
                   pao->ChannelObject[channel].MemoryVaAdj.ptr +
                        pao->ChannelObject[channel].RqDescOffset.uint,
                   RecvPackets,
                   RecvPacketSize,
                   1 );

    /*************************************************************************
        Setup the device ccb with physical addresses.  Does not require a 
        doorbell address since the adapter has its own reference to the
        doorbell via local register access on the adapter.

        6/22/2001 tjb - Added retry-code to handle the case when we have
                        mapped the ccb table entry at the adapter and for
                        some reason the adapter experienced a reset event.
                        All we need to do if a reset event has been detected
                        is clear the fifo handle reset fields and try to map
                        the ccb table entry again to ensure that the newly
                        booted firmware on the target adapter see's the fifo
                        addresses for this channel.
    **************************************************************************/

_mapccbtableentry:
    DBG("Calling CpqCiCcbSetup for Physical Kernel Mode Addresses . . .\n");
    CpqCiCcbSetup( &ccbtbl[channel],
                   channel,
                   CPQCINULL,
                   pao->ChannelObject[channel].MemoryPaAdj.ptr +
                        pao->ChannelObject[channel].SqFifoOffset.uint,
                   pao->ChannelObject[channel].MemoryPaAdj.ptr +
                        pao->ChannelObject[channel].SqDescOffset.uint,
                   SendPackets,
                   SendPacketSize,
                   1,
                   pao->ChannelObject[channel].MemoryPaAdj.ptr +
                        pao->ChannelObject[channel].RqFifoOffset.uint,
                   pao->ChannelObject[channel].MemoryPaAdj.ptr +
                        pao->ChannelObject[channel].RqDescOffset.uint,
                   RecvPackets,
                   RecvPacketSize,
                   1 );

    //  If a reset condition is detected after we just mapped the hardware's
    //  CCB table entry for this channel, implement a back off until the reset
    //  condition has been cleared (by the ISR/DPC) before attempting to re-map
    //  the already allocated addresses to the hardware's CCB table entry for
    //  this channel.
    if( CpqCiAdapterReset(pregs) || CpqCiChannelReset(driverccb) )
    {
        //  The adapter's reset condition will be dealt with by the ISR/DPC
        //  handler.  We need to deal with the memory based reset fields in the
        //  fifo handles.
        while( CpqCiAdapterReset(pregs) || CpqCiChannelReset(driverccb) )
        {
            DBG("Reset Sq/RqFifo\n");
            CpqCiFifoBar2FifoHandle(driverccb->SqFifo.ptr)->reset.uint = 0;
            CpqCiFifoBar2FifoHandle(driverccb->RqFifo.ptr)->reset.uint = 0;
        }

        goto _mapccbtableentry;
    }

    /*************************************************************************
        Map the host memory region into the callers address space.  If the
        caller has originated from user mode, then the mapped ccb will contain
        the virtual address information already created for the driver ccb.
    **************************************************************************/

    DBG("Calling MapConnection\n");
    error = 
    MapConnection( Context,
                   Adapter,
                   channel,
                   AddressMode,
                   &pao->ConnectionObject[connection].MappedCcbCopy,
                   &pao->ConnectionObject[connection].MappedMemoryHandle,
                   &pao->ConnectionObject[connection].MappedDoorbellHandle );

    if( error != CpqCiSuccess )
        goto exitProc;
    
    /*************************************************************************
        Construct the key used to grant shared access to this channel.  This
        key will be derived from the driver's ccb information.
    **************************************************************************/

    ConstructKey(&pao->ChannelObject[channel].Key, driverccb);

    /*************************************************************************
        Update the caller's buffers if supplied.
    **************************************************************************/

    DBG("Update the callers buffers if supplied\n");

    if( pKey )
        CpqCiCopyBuffer( pKey,
                         &pao->ChannelObject[channel].Key,
                         sizeof(CPQCIKEY) );

    if( CallersCcb )
        CpqCiCopyBuffer( CallersCcb,
                         &pao->ConnectionObject[connection].MappedCcbCopy,
                         sizeof(CPQCICCB) );

    /*************************************************************************
        Place descriptors on send list with 0-length.  A 0-length descriptor
        instructs the HW to immediately turn the packet around and mark it as
        "free" without generating a response.
    **************************************************************************/

    DBG("Place descriptors on send list with 0 length\n");
    packetsize = 0;
    for( packetid = 0; packetid < SendPackets; packetid++ )
    {
    	DBG("packetid = %d driverccb = %p\n", packetid, driverccb);
        CpqCiPacketEnqueue(driverccb, CpqCiSendQ, packetid, packetsize);
        CpqCiDoorbellSet(driverccb);
    }

    //  Place descriptors on recv list.  The HW uses these descriptors
    //  to generate responses for this channel.
    DBG("Place descriptors on recv list \n");
    packetsize = CpqCiDescMemorySize( driverccb->RqCtrl, 1 );
    for( packetid = 0; packetid < RecvPackets; packetid++ ) 
    {
    	DBG("packetid = %d driverccb = %p\n", packetid, driverccb);
        CpqCiPacketEnqueue(driverccb, CpqCiRecvQ, packetid, packetsize);
    }

    CpqCiDoorbellClr(driverccb);

    /*************************************************************************
        Make sure the target firmware has marked the send packets as "free".
    **************************************************************************/

    DBG("Make sure the target fw has marked send packes as free\n");
    //  09/14/2001 tjb: (b)
    //  10/17/2001 tjb: (a)
    for( attempts = 0x200000; attempts; attempts-- )
    {
        if( CpqCiPacketDequeue( driverccb,
                                CpqCiSendQ,
                                &packetid,
                                CPQCINULL,
                                CPQCINULL  ) )
        {
		DBG("PacketID = %d\n", packetid);
            break;
        }

        //  09/14/2001 tjb: (b)
        CpqCiDrvOsStall( CPQCINULL, 1 );

    }

    //  Remaining attempts indicate that we pulled off free send packet.
    //  Put this packet id back onto the send list.
    if( attempts )
    {
	DBG("%d attempts to CpqCiPacketDequeue\n", (0x200000 - attempts));
	DBG("Put packet back into send queue\n");
        CpqCiPacketEnqueue(driverccb, CpqCiSendQ, packetid, 0);
        CpqCiDoorbellSet(driverccb);
    }
    else
    {
	DBG("ERROR channel not responding\n");
        error = CpqCiErrChannelNotResponding;
        goto exitProc;
    }

    /*************************************************************************
        Leave the routine and finish appropriately.
    **************************************************************************/

    pao->ChannelObject[channel].ReferenceCount.uint = 1;
    pao->ConnectionObject[connection].Context = NULL;
    error = CpqCiSuccess;

#ifdef DEBUG
    //Dump out the current driverccb.
    DBG("Dump CCB\n");
    DumpCpqCiCcb(driverccb);
#endif

exitProc:

    if( error != CpqCiSuccess )
    {
        DBG("Exiting with FAILURE!!!\n\n");
        //  A valid connection number requires us to clear the channel object
        //  and to place the connection number back onto its fifo.
        if( connection != CPQCI_INVALID_CONNECTION )
        {
            CpqCiDrvOsFree( 
                Context, 
                pao->ConnectionObject[connection].MappedMemoryHandle
                );
            CpqCiDrvOsFree( 
                Context, 
                pao->ConnectionObject[connection].MappedDoorbellHandle
                );
            CpqCiFillBuffer( 
                &pao->ConnectionObject[connection],
                0,
                sizeof(CPQCICONNECTIONOBJECT)
                );
            CpqCiFifoEnqueue(pao->QConnectionObjs, connection);
        }

        //  A valid channel number requires us to clear the channel object
        //  and place the channel number back onto its priority fifo.
        if( channel != CPQCI_INVALID_CHANNEL )
        {
            CpqCiDrvOsFree(Context, pao->ChannelObject[channel].MemoryHandle);
            CpqCiFillBuffer( 
                &pao->ChannelObject[channel],
                0,
                sizeof(CPQCICHANNELOBJECT)
                );
            //  09/14/2001 tjb: (a)
            CpqCiFillBuffer( &ccbtbl[channel], 0, ((CPQCICCB *)(0))->Context );
            CpqCiFifoEnqueue(qpriority, channel);
        }
    }

    if( pChannel )
        *pChannel = channel;

    if( pConnection )
        *pConnection = connection;

    DBG("\n CpqCiDrvCreateChannel EXIT (0x%x)\n", error);
        
    return error;
}

CPQCISTATUS
CpqCiDrvOpenChannel(
    CPQCICONTEXT Context,
    CPQCIADAPTER Adapter,
    CPQCIMODE AddressMode,
    CPQCICCB *CallersCcb,
    CPQCIKEY *pKey,
    CPQCICONNECTION *pConnection,
    CPQCICHANNEL *pChannel
    )

/*++

Routine Description:

    This routine creates a connection to an already existing channel.

Arguments:

    Context - The caller's supplied context information required for some 
        OS dependent functions such as the CpqCiDrvOsXx() memory management
        routines.

    Adapter - Target adapter object.

    AddressMode - The requested address mode to map the connection address
        information.  Valid values for this field are the following.

            CpqCiUserMode
            CpqCiKernelMode

    CallersCcb - Pointer to caller's CCB filled in with context specific 
        address information to operate on the channel backing the connection.
        This buffer will only get filled if not NULL.

    pKey - Pointer to the caller's key buffer that will be used to determine
        if the caller has valid access to the requested channel.  The key 
        contains information about the channel as it was created.

    pConnection - Location to place the connection created for this channel.
        The contents of this location are only valid when this routine returns
        CpqCiSuccess.

    pChannel - Location to place the 0-based channel number associated with 
        this connection.

Return Value:

    The result of the operation.
    
Caveats:

    none

--*/
{
    CPQCIADAPTEROBJECT *pao = Adapter;
    CPQCIBYTE *pregs = CpqCiDrvOsGetVA(Context, pao->MmioRegistersHandle);
    CPQCICONNECTION connection = CPQCI_INVALID_CONNECTION;
    CPQCICHANNEL channel = CPQCI_INVALID_CHANNEL;
    CPQCISTATUS error = CpqCiSuccess;
    CPQCIBYTE keyvalue;
    CPQCICCB key2ccb;
    CPQCIBYTE *pcmp1;
    CPQCIBYTE *pcmp2;
    CPQCISIZE size;
    vu64_t *pref;
    u64_t ref;

    /*************************************************************************
        Convert the key into a CCB.
    **************************************************************************/

    if( !pKey )
    {
        error = CpqCiErrNoKeySpecified;
        goto exitProc;
    }

    ConstructKey( &key2ccb, pKey );

    /*************************************************************************
        Extract the target channel number so we can obtain the correct key
        buffer to validate the caller supplied key.  Treat an invalid channel
        number the same as a mismatch on the key.
    **************************************************************************/

    channel = CpqCiChannelNumberGet( &key2ccb );

    if( channel >= pao->MaxChannels )
    {
        channel = CPQCI_INVALID_CHANNEL;
        error = CpqCiErrAccessDenied;
        goto exitProc;
    }

    /*************************************************************************
        Validate the caller's key with the created key for the target channel.
        Valid keys do not contain all zeros.
    **************************************************************************/

    keyvalue = 0;
    pcmp1 = (CPQCIBUFFER)pKey;
    pcmp2 = (CPQCIBUFFER)&pao->ChannelObject[channel].Key;
    size  = CPQCI_KEY_SIZE;

    while( size-- )
    {
        keyvalue |= *pcmp2;
        if( *pcmp1++ != *pcmp2++ )
        {
            channel = CPQCI_INVALID_CHANNEL;
            error = CpqCiErrAccessDenied;
            goto exitProc;
        }
    }

    //  The OR of all key bits better not be 0!
    if( !keyvalue )
    {
        channel = CPQCI_INVALID_CHANNEL;
        error = CpqCiErrAccessDenied;
        goto exitProc;
    }

    /*************************************************************************
        Try to obtain a connection object number for the requested channel.
    **************************************************************************/

    if( !CpqCiFifoDequeue(pao->QConnectionObjs, (CPQCIENTRY *)&connection) )
    {
        connection = CPQCI_INVALID_CONNECTION;
        error = CpqCiErrNoMoreConnections;
        goto exitProc;
    }

    /*************************************************************************
        Try to claim a reference to the target channel by bumping the 
        reference count.  Keep in mind that once we obtain a reference, we
        are responsible for removing the reference appropriately if an error
        is encountered during a memory mapping operation.  This means that
        we are responsible for adding the channel back to the free channel
        fifo if we remove our reference and happen to be the last reference
        to this channel.
    **************************************************************************/

    pref = (vu64_t *)&pao->ChannelObject[channel].ReferenceCount;

    CpqCiEnterLoop

        //  Take a snapshot of the current reference count.
        ref = *pref;

        //  We can only proceed if the current reference has not reached 0.
        if( ref )
        {
            //  6/22/2001 tjb - Add code to detect a reset condition and 
            //                  fail the open attempt.
            if( CpqCiAdapterReset(pregs) )
            {
                error = CpqCiErrChannelDisconnected;
                goto exitProc;
            }

            //  Try to add our reference to this channel object.
            if( CpqCiCas(pref, ref, ref+1) )
            {
                //  Map the caller's address information.
                error = 
                MapConnection(
                    Context,
                    Adapter,
                    channel,
                    AddressMode,
                    &pao->ConnectionObject[connection].MappedCcbCopy,
                    &pao->ConnectionObject[connection].MappedMemoryHandle,
                    &pao->ConnectionObject[connection].MappedDoorbellHandle 
                    );

		pao->ConnectionObject[connection].Context = NULL;

                if( error != CpqCiSuccess )
                {
                    //  Close the connection and decrement our reference.
                    //  CpqCiDrvCloseChannel() will perform any required
                    //  cleanup.
                    CpqCiDrvCloseChannel( Context,
                                          Adapter,
                                          connection,
                                          &channel );
                    goto exitProc;
                }

                break;  // done!
            }
        }
        else
        {
            //  The channel is marked for deletion and cannot be accessed.
            channel = CPQCI_INVALID_CHANNEL;
            error = CpqCiErrChannelDeleted;
            goto exitProc;
        }

    CpqCiLeaveLoop

    /*************************************************************************
        Map the connection's ccb address information into the caller's ccb.
    **************************************************************************/

    if( CallersCcb )
        CpqCiCopyBuffer( CallersCcb,
                         &pao->ConnectionObject[connection].MappedCcbCopy,
                         sizeof(CPQCICCB) );

exitProc:

    if( pChannel )
        *pChannel = channel;

    if( pConnection )
        *pConnection = connection;
        
    return error;
}

CPQCISTATUS
CpqCiDrvCloseChannel(
    CPQCICONTEXT Context,
    CPQCIADAPTER Adapter,
    CPQCICONNECTION Connection,
    CPQCICHANNEL *pChannel
    )

/*++

Routine Description:

    This routine closes a channel connection.

Arguments:

    Context - The caller's supplied context information required 
        for some OS dependent functions such as the CpqCiDrvOsXx() 
        memory management routines.

    Adapter - Target adapter object.

    Connection - The connection to close.

    pChannel - Location to place the target channel number being closed.

Return Value:

    The result of the operation.
    
Caveats:

    This routine will always close the connection, however, there may be
    other connections referring to the target channel.  This routine will
    detect outstanding references to this channel and if detected, just 
    delete the caller's context information without deleting the resources 
    backing the channel.  If the caller is the last reference to the channel,
    then the caller's resources are unmapped and the channels resources are 
    reclaimed.

--*/
{
    CPQCIADAPTEROBJECT *pao = Adapter;
    CPQCISTATUS error = CpqCiSuccess;
    CPQCICHANNEL channel = CPQCI_INVALID_CHANNEL;
    CPQCICCB *ccbtbl = CPQCINULL;
    CPQCICCB *driverccb = CPQCINULL;
    CPQCICCB mappedccb;
    CPQCIBOOL connectionFreed = 1;
    vu64_t *pref;
    u64_t ref;

    /*************************************************************************
        We're done if there is no adapter object.
    **************************************************************************/

    if( !pao )
        goto exitProc;

    /*************************************************************************
        Validate the connection.
    **************************************************************************/

    if( Connection >= pao->MaxConnections )
    {
        error = CpqCiErrBadConnection;
        goto exitProc;
    }

    /*************************************************************************
        Extract the relevant fields.  Copy the mapped ccb into a local ccb
        so we can reconcile the connection object.
    **************************************************************************/

    CpqCiCopyBuffer( &mappedccb,
                     &pao->ConnectionObject[Connection].MappedCcbCopy,
                     ((CPQCICCB *)(0))->Context );

    channel = CpqCiChannelNumberGet( &mappedccb );
    driverccb = &pao->ChannelObject[channel].DriverCcb;
    ccbtbl = CpqCiDrvOsGetVA(Context, pao->MmioCcbTableHandle);

    /*************************************************************************
        Unmap our context information, clear the connection object and place
        the connection object number onto its fifo queue.
    **************************************************************************/


    ref = 0;
    CpqCiOsGetRefCount(pao->ConnectionObject[Connection].Context, (CPQCITYPEQUAD*) &ref);
    DBG(">>> Connection %d obj %p ref %ld\n", Connection, &pao->ConnectionObject[Connection], ref);

    if( ref <= 3) 
    {
	    CPQCICONNECTIONOBJECT* ConnectionObject = &pao->ConnectionObject[Connection];
            //  Fill connection object if we were the last reference.
	    DBG(">>> Destroy Connection %d and channel %d\n", Connection, channel);

	    connectionFreed = 1;

	    DBG(">>> Context %p Unmap %p\n", CpqCiGetCurrentTask() , ConnectionObject->MappedMemoryHandle);
	    CpqCiDrvOsFree( 
			Context,
			ConnectionObject->MappedMemoryHandle
	    );
	    DBG(">>> Context %p Unmap %p\n", CpqCiGetCurrentTask(), ConnectionObject->MappedDoorbellHandle);
	    CpqCiDrvOsFree(
			Context,
			ConnectionObject->MappedDoorbellHandle
	    );

	    CpqCiFillBuffer(
		ConnectionObject,
		0,
		sizeof(CPQCICONNECTIONOBJECT)
	    );


    } else {
	    DBG(">>> DO NOT destroy Connection %d nor channel %d\n", Connection, channel);
	    connectionFreed = 0;
            error = CpqCiErrConnectionStillInuse;
            goto exitProc;
    }

//    CpqCiFifoEnqueue(pao->QConnectionObjs, Connection); //10/11/2001 tjb: (a)

    /*************************************************************************
        Remove our reference to the channel.
    **************************************************************************/

    pref = (vu64_t *)&pao->ChannelObject[channel].ReferenceCount;

    CpqCiEnterLoop

        //  Take a snapshot of the current reference count.
        ref = *pref;

        DBG(">>> Channel %d obj %p ref %ld\n", channel, &pao->ChannelObject[channel], ref);
        //  Try to remove our reference to this channel object.
        if( CpqCiCas(pref, ref, ref-1) )
        {
            //  Free channel resources if we were the last reference.
            if( ref == 1)
            {
                u8_t *pregs = CpqCiDrvOsGetVA(Context, pao->MmioRegistersHandle);
                CPQCIFIFOBAR qpriority = pao->ChannelObject[channel].QPriority;
                CPQCIHANDLE memoryhandle;
                CPQCILENGTH retries;

                //  Stop device interrupt chatter and clear pending 
                //  doorbells.
                DBG(">>> Channel = %d, cctbl = %x\n" , channel, ccbtbl);
                CpqCiDoorbellMask(&ccbtbl[channel]);
                DBG(">>> DoorbellClr\n");
                CpqCiDoorbellClr(driverccb);

                DBG(">>> stop adapter\n");
                //  Stop the adapter from processing this channel for both
                //  sends and receives.
                CpqCiCtrlClearGo(ccbtbl[channel].SqCtrl);
                CpqCiCtrlClearGo(ccbtbl[channel].RqCtrl);

                //  Make sure the adapter goes inactive for this channel.
                //  Currently 10 seconds.
                for( retries = 10 * 100000; retries; retries-- )
                {
                    //  Hit the send fifo doorbell to get the adapter's
                    //  attention towards our cleared go bits.  It is
                    //  possible that the adapter is idle for this
                    //  channel and therefore, would not want to timeout
                    //  on the close.
                    CpqCiDoorbellSet(driverccb);

                    //  Stall before checking adapter.
                    CpqCiDrvOsStall(Context, 10); // stall in uSeconds

                    //  Check to see if adapter has gone inactive for
                    //  this channel.
                    if( !CpqCiCtrlTestActive(ccbtbl[channel].SqCtrl) &&
                        !CpqCiCtrlTestActive(ccbtbl[channel].RqCtrl) )
                        break; // channel has been stopped!

                }

		if (retries <= 0)
			DBG("CpqCiDrvCloseChannel:  retries <= 0\n");

                //  Clear the target ccb for this channel at the adapter.
                //  Make sure not to blow away the context region.
                CpqCiFillBuffer( &ccbtbl[channel],
                                 0,
                                 ((CPQCICCB *)(0))->Context );

                //  Extract the memory handle we will need to wack before
                //  clearing out the channel object.
                memoryhandle = pao->ChannelObject[channel].MemoryHandle;

                //  Before clearing the channel object, it is important to
                //  clear the SqFifo address with a single ATOMIC write.
                //  The fill operation may fill bytes rather than QWORDS
                //  which could cause a partial read of the SqFifo field
                //  by the completion handler and hence, reap an invalid 
                //  address for the fifo handle.
                pao->ChannelObject[channel].DriverCcb.SqFifo.uint = 0;

                //  Hold off wacking the buffer until the reset handler
                //  clears the reset event pending at the adapter.  This
                //  ensures that the ISR/DPC routine has finished touching
                //  the channel memory.  Note that this wait feature only
                //  occurs for the case the adapter has encountered a reset
                //  condition during normal operation.
                while( CpqCiAdapterReset(pregs) ) {
			DBG(">>> Adapter in Reset\n");
		}

                DBG(">>> FillBuffer\n");
                //  Clear the channel object and return the channel to its
                //  target priority queue.
                CpqCiFillBuffer( &pao->ChannelObject[channel],
                                 0,
                                 sizeof(CPQCICHANNELOBJECT) );

                //  Wack the host memory buffer.
                DBG(">>> Whack host memory buffer Context %p mh %p\n", Context, memoryhandle);
                CpqCiDrvOsFree( Context, memoryhandle );

                //  Place the channel number back onto the free list for other
                //  channel customers.
		DBG(">>> Enqueue prior %p channel %d\n", qpriority, channel);
		CpqCiFifoEnqueue(qpriority, channel);

                break;  // done!

            }
            else
            {
                error = CpqCiErrChannelStillInuse;
                goto exitProc;
            }
        }

    CpqCiLeaveLoop

    error = CpqCiSuccess;

exitProc:

    //  10/11/2001 tjb: (a)
    if( pao && (error != CpqCiErrBadConnection) && connectionFreed ) {
	DBG(">>> Connection %d gets put back in the free pool\n", Connection);
        CpqCiFifoEnqueue(pao->QConnectionObjs, Connection);
    }

    if( pChannel )
        *pChannel = channel;

    return error;
}

CPQCILENGTH
CpqCiDrvSend(
    CPQCICONTEXT Context,
    CPQCIADAPTER Adapter,
    CPQCICONNECTION Connection,
    CPQCIBUFFER Buffer,
    CPQCILENGTH Length,
    CPQCISTATUS *pError
    )

/*++

Routine Description:

    This routine posts a message onto the send fifo of the target channel.

Arguments:

    Context - The caller's supplied context information required for some 
        OS-dependent functions such as the CpqCiDrvOsXx() memory management
        routines.

    Adapter - Target adapter object.

    Connection - Target connection containing channel information.

    Buffer - The caller supplied message buffer that will get posted onto
        the send fifo.  The contents of this buffer are immediately available
        to the caller upon return of this function.

    Length - The number of bytes to post to the send fifo for the target
        channel.

    pError - Location to place the result of the operation.  If this value is 
        passed as NULL (0), then no error will be returned.

Return Value:

    The actual number of bytes sent.
    
Caveats:

    This routine does not support blocking.

--*/
{
    CPQCICCB *ccb = &Adapter->ConnectionObject[Connection].MappedCcbCopy;
    CPQCICHANNEL channel = CpqCiChannelNumberGet(ccb);
    CPQCISTATUS error = CpqCiSuccess;
    CPQCIPACKETID pid;
    CPQCIPACKET pkt;
    CPQCILENGTH len=0; // 08/22/2002 tjb: (a)

    //  We always want to use the kernel mode context of the ccb.  This
    //  will keep us off of a possible user-mapped ccb context.
    ccb = &Adapter->ChannelObject[channel].DriverCcb;

    //  We report a possible reset condition before any send attempt.
    if( CpqCiChannelReset(ccb) )
    {
        error = CpqCiErrChannelDisconnected;
        goto exitProc;
    }

    //  Obtain a free send packet from the send fifo.  If there are no
    //  more send packets, then logically the send FIFO is full.
    if( !CpqCiPacketDequeue(ccb, CpqCiSendQ, &pid, &len, &pkt) )
    {
        error = CpqCiErrFifoFull;
        goto exitProc;
    }

    //  Check the caller's buffer length against the length of the 
    //  send packet.  If the caller's buffer is larger than the size
    //  of the send packet, then only copy what we can into the send
    //  packet and indicate a partial transfer.
    if( Length <= len )
        len = Length;
    else
        error = CpqCiErrPartialTransfer;

    //  Copy the caller's message data into the send packet.
    CpqCiCopyBuffer( pkt, Buffer, len );

    //  Post the send packet to the send FIFO and ring the doorbell.
    CpqCiPacketEnqueue(ccb, CpqCiSendQ, pid, len);
    CpqCiDoorbellSet( ccb );

exitProc:
    if( pError )
        *pError = error;

    return len;
}

CPQCILENGTH
CpqCiDrvRecv(
    CPQCICONTEXT Context,
    CPQCIADAPTER Adapter,
    CPQCICONNECTION Connection,
    CPQCIBUFFER Buffer,
    CPQCILENGTH Length,
    CPQCISTATUS *pError
    )

/*++

Routine Description:

    This routine reaps a message off of the receive fifo of the target channel.

Arguments:

    Context - The caller's supplied context information required for some 
        OS-dependent functions such as the CpqCiDrvOsXx() memory management
        routines.

    Adapter - Target adapter object.

    Connection - Target connection containing channel information.

    Buffer - The caller supplied message buffer filled in with the next
        available message from the receive fifo.

    Length - The size in bytes of the caller supplied message buffer.

    pError - Location to place the result of the operation.

Return Value:

    The actual number of bytes received.
    
Caveats:

    This routine does not support blocking.

--*/
{
    CPQCICCB *ccb = &Adapter->ConnectionObject[Connection].MappedCcbCopy;
    CPQCICHANNEL channel = CpqCiChannelNumberGet(ccb);
    CPQCISTATUS error = CpqCiSuccess;
    CPQCIPACKETID pid;
    CPQCIPACKET pkt;
    CPQCILENGTH len=0; // 08/22/2002 tjb: (a)

    //  We always want to use the kernel mode context of the ccb.  This
    //  will keep us off of a possible user-mapped ccb context.
    ccb = &Adapter->ChannelObject[channel].DriverCcb;

    //  We report a possible reset condition before any receive attempt.
    if( CpqCiChannelReset(ccb) )
    {
        error = CpqCiErrChannelDisconnected;
        goto exitProc;
    }

    //  Obtain a receive packet from the receive fifo.  If there are no
    //  receive packets, then logically the receive FIFO is empty.
    if( !CpqCiPacketDequeue( ccb, CpqCiRecvQ, &pid, &len, &pkt ) )
    {
        error = CpqCiErrFifoEmpty;
        goto exitProc;
    }

    //  At this point, we have a receive packet and need to copy it
    //  into the caller's receive buffer.  Make sure the caller's buffer
    //  is large enough to receive the data, otherwise indicate a partial
    //  transfer.
    if( Length < len )
    {
        len = Length;
//        error = CpqCiErrPartialTransfer; //10/11/2001 tjb: (b)
    }

    //  Copy the packet data into the caller's buffer and place the packet
    //  back onto the receive fifo.
    CpqCiCopyBuffer( Buffer, pkt, len );
    CpqCiPacketEnqueue(ccb, CpqCiRecvQ, pid, CpqCiDescMemorySize(ccb->RqCtrl,1));

exitProc:
    if( pError )
        *pError = error;

    return len;
}

CPQCISTATUS
CpqCiDrvRegisterWaiter(
    CPQCICONTEXT Context,
    CPQCIADAPTER Adapter,
    CPQCICONNECTION Connection
    )

/*++

Routine Description:

    This routine registers a thread as a waiter on a receive notification.

Arguments:

    Context - The caller's supplied context information required for some
        OS-dependent functions such as the CpqCiDrvOsXx() memory management
        routines.

    Adapter - Target adapter object.

    Connection - Target connection.

Return Value:

    The result of the operation.
    
Caveats:

    None.

--*/
{
    CPQCICCB *ccbtbl = CpqCiDrvOsGetVA(Context, Adapter->MmioCcbTableHandle);
    CPQCIBYTE *pregs = CpqCiDrvOsGetVA(Context, Adapter->MmioRegistersHandle);
    CPQCICCB *ccb = &Adapter->ConnectionObject[Connection].MappedCcbCopy;
    CPQCICHANNEL channel = CpqCiChannelNumberGet(ccb);
    vu64_t *pwaiters = (void *)&Adapter->ChannelObject[channel].WaiterCount;
    u64_t waiters;

    //  We always want to use the kernel mode context of the ccb.  This
    //  will keep us off of a possible user-mapped ccb context.
    ccb = &Adapter->ChannelObject[channel].DriverCcb;

    DBG("DrvRegisterWaiter pregs %p, ccb %p\n", pregs, ccb);

    CpqCiDoorbellUnmask( &ccbtbl[channel] );

    CpqCiEnterLoop
        waiters = *pwaiters;
        if(CpqCiCas(pwaiters, waiters, /*waiters+(02/18/2002 tjb: (a))*/1))
            goto exitProc;
    CpqCiLeaveLoop

exitProc:
    if( CpqCiAdapterReset(pregs) || CpqCiChannelReset(ccb) )
        return CpqCiErrChannelDisconnected;
    else
        return CpqCiSuccess;

}

CPQCISTATUS
CpqCiDrvUnregisterWaiter(
    CPQCICONTEXT Context,
    CPQCIADAPTER Adapter,
    CPQCICHANNEL Channel
    )

/*++

Routine Description:

    This routine unregisters a thread as a waiter on a receive notification.

Arguments:

    Context - The caller's supplied context information required for some
        OS-dependent functions such as the CpqCiDrvOsXx() memory management
        routines.

    Adapter - Target adapter object.

    Channel - Target channel.

Return Value:

    The result of the operation.
    
Caveats:

    None.

--*/
{
    CPQCICCB *ccbtbl = CpqCiDrvOsGetVA(Context, Adapter->MmioCcbTableHandle);
    vu64_t *pwaiters = (void *)&Adapter->ChannelObject[Channel].WaiterCount;
    u64_t waiters;

    CpqCiDoorbellMask( &ccbtbl[Channel] );

    CpqCiEnterLoop
        waiters = *pwaiters;
        if( waiters == 0 )
            return CpqCiErrNoMoreWaiters;
        if( CpqCiCas(pwaiters, waiters, waiters-1) )
            return CpqCiSuccess;
    CpqCiLeaveLoop
}

CPQCISTATUS
CpqCiDrvGetPendingChannels(
    CPQCICONTEXT Context,
    CPQCIADAPTER Adapter,
    CPQCICHANNEL *pChannel,
    CPQCICHANNELS *pChannelList,
    CPQCICHANNELS *pDeferredList
    )

/*++

Routine Description:

    This routine retrieves the next pending channel based on priority.

Arguments:

    Context - The caller's supplied context information required for some OS
        dependent functions such as the CpqCiDrvOsXx() memory management 
        routines.

    Adapter - Target adapter object.

    pChannel - Location to place the 0-based channel number pending service 
        on the outbound RECV FIFO.

    pChannelList - Location to receive and update the pending channel list.  
        If the contents of this location is 0, then this routine will retrieve
        a new set of pending channels from the device outbound doorbell 
        register and clear the pending bits.  Otherwise, this routine will
        process the channels already found in the channel list.

    pDeferredList - Location containing a "snapshot" of channels that
        were deferred due to some event such as more waiters on this channel.

Return Value:

    The result of the operation.  The content of pChannel is only valid if this
    routine returns CpqCiSuccess.
    
Caveats:

    This routine attempts to service the outbound doorbell register based 
    on channel priority.  A check is always done to see if the HIGHPRIORITY
    channels (23:20) are pending before running the NORMALPRIORITY channel 
    list.

--*/
{
    //  This list does not reside on the stack.  It resides in a
    //  private data region only visible by this routine.
    static u8_t Channels[] =
    {
        255,    // no channels pending
        00,
        01,01,
        02,02,02,02,
        03,03,03,03,03,03,03,03,
        04,04,04,04,04,04,04,04,04,04,04,04,04,04,04,04,
        05,05,05,05,05,05,05,05,05,05,05,05,05,05,05,05,
        05,05,05,05,05,05,05,05,05,05,05,05,05,05,05,05,
        06,06,06,06,06,06,06,06,06,06,06,06,06,06,06,06,
        06,06,06,06,06,06,06,06,06,06,06,06,06,06,06,06,
        06,06,06,06,06,06,06,06,06,06,06,06,06,06,06,06,
        06,06,06,06,06,06,06,06,06,06,06,06,06,06,06,06,
        07,07,07,07,07,07,07,07,07,07,07,07,07,07,07,07,
        07,07,07,07,07,07,07,07,07,07,07,07,07,07,07,07,
        07,07,07,07,07,07,07,07,07,07,07,07,07,07,07,07,
        07,07,07,07,07,07,07,07,07,07,07,07,07,07,07,07,
        07,07,07,07,07,07,07,07,07,07,07,07,07,07,07,07,
        07,07,07,07,07,07,07,07,07,07,07,07,07,07,07,07,
        07,07,07,07,07,07,07,07,07,07,07,07,07,07,07,07,
        07,07,07,07,07,07,07,07,07,07,07,07,07,07,07,07
    };

    CPQCIADAPTEROBJECT *pao = Adapter;
    CPQCISTATUS error = CpqCiErrNoPendingChannels;
    CPQCIBYTE *pregs = CpqCiDrvOsGetVA(Context, pao->MmioRegistersHandle);
    vu32_t *pdoorbell = (void *)&pregs[CPQCI_REG_DBOUTBOUND];
    u32_t list;

    //  Take a snapshot of pending channels.  Do not handle channels 31:24.
    //
    //  Caveat: The bits for each mask contain the supported channels for 
    //          the specific priority.  For example, if we wish to support 
    //          only critical priority channels 26:24, we would change the 
    //          critical mask from 0x00000000 to 0x07000000.
    //
    //  6/22/2001 tjb - Added reset-notification handling by setting bit 26
    //                  in the critical mask constant to support the RESET
    //                  bit assertion borrowing the outbound doorbell's bit 26.
    //                                       critical     high       normal
    list = (*pdoorbell | *pDeferredList) & (0x04000000|0x00F00000|0x000FFFFF);

    //  Test for the reset condition.  If a reset was detected, then we will
    //  need to spin down each channel and set the reset fields within each
    //  channel's valid host memory region and then set the channel bit map
    //  to indicate all channel's pending to wake up any waiters on channels
    //  that have not yet responded.
    if( list & 0x04000000 )
    {
        CPQCICHANNEL channel;
        CPQCITYPEQUAD sqfifo;
        CPQCITYPEQUAD rqfifo;

        for( channel = 0; channel < pao->MaxChannels; channel++ )
        {
	    DBG_INTR("channel %d\n", channel);
            sqfifo.ptr = pao->ChannelObject[channel].DriverCcb.SqFifo.ptr;
            rqfifo.ptr = pao->ChannelObject[channel].DriverCcb.RqFifo.ptr;

            // 09/10/2001 tjb: (a)
            if( sqfifo.ptr )
                CpqCiFifoBar2FifoHandle(sqfifo.ptr)->reset.uint = 1;

            // 09/10/2001 tjb: (a)
            if( rqfifo.ptr )
                CpqCiFifoBar2FifoHandle(rqfifo.ptr)->reset.uint = 1;
        }

        //  Force completion handler to wake up all waiters on all channels.
        *pChannelList = 0x00FFFFFF;
    }

    DBG_INTR("Blend new pending events\n");
    //  Blend in all of the new pending events if the current list is 0.
    if( !*pChannelList )
    {
        if( list ) 
            *pChannelList = list;
        else
            //  No pending events found.  We're done!
            goto exitProc;
    }

    //  Always blend in critical priority events.  This is done because
    //  there is no fairness in channel processing for critical channels.
    //  The priority assignment is strict left-to-right where channel 31
    //  is the highest and channel 24 is the lowest within this priority
    //  class.
    //
    //  Caveat: We do not currently receive critical events as seen by the
    //          mask applied to the result of the doorbell read operation.
    DBG_INTR("Blend critical priority events\n");
    *pChannelList |= (list & 0xFF000000);

    //  There are events remaining to be serviced found in the caller's 
    //  copy of pending events.  If the caller's channel list has no
    //  pending high priority events, blend in any new ones we found from
    //  the outbound doorbell register.  We need to do this inorder to
    //  ensure that higher priority channels get serviced before normal
    //  priority channels.
    DBG_INTR("Blend high priority events\n");
    if( !(*pChannelList & 0x00F00000) )
        *pChannelList |= (list & 0x00F00000);

    //  Clear the hardware of pending events that we have already seen
    //  including the deferred channel list bits.
    DBG_INTR("Doorbell\n");
    *pdoorbell = *pChannelList;
    DBG_INTR("Deferred List\n");
    *pDeferredList &= ~*pChannelList;

    DBG_INTR("Check 31:24\n");
    //  Check first for pending critical priority channels 31:24.
    list = (*pChannelList & 0xFF000000)>>CPQCI_BASE_CRITICALPRIORITY_CHANNEL;
    if( list )
    {
        //  Found one.  Extract it and reset bit in caller's list.
        *pChannel = (CPQCICHANNEL)
            (Channels[ list ] + CPQCI_BASE_CRITICALPRIORITY_CHANNEL);
        *pChannelList &= ~(1 << *pChannel);
        error = CpqCiSuccess;
        goto exitProc;
    }

    DBG_INTR("Check 23:20\n");
    //  Check next for pending high priority channels 23:20.
    list = (*pChannelList & 0x00F00000)>>CPQCI_BASE_HIGHPRIORITY_CHANNEL;
    if( list )
    {
        //  Found one.  Extract it and reset bit in caller's list.
        *pChannel = (CPQCICHANNEL)
            (Channels[ list ] + CPQCI_BASE_HIGHPRIORITY_CHANNEL);
        *pChannelList &= ~(1 << *pChannel);
        error = CpqCiSuccess;
        goto exitProc;
    }

    DBG_INTR("Check 19:16\n");
    //  Check for normal priority channels 19:16.
    list = (*pChannelList & 0x000F0000)>>(CPQCI_BASE_NORMALPRIORITY_CHANNEL+16);
    if( list )
    {
        //  Found one.  Extract it and reset bit in caller's list.
        *pChannel = (CPQCICHANNEL)
            (Channels[ list ] + (CPQCI_BASE_NORMALPRIORITY_CHANNEL+16));
        *pChannelList &= ~(1 << *pChannel);
        error = CpqCiSuccess;
        goto exitProc;
    }

    DBG_INTR("Check 15:8\n");
    //  Check for normal priority channels 15:8.
    list = (*pChannelList & 0x0000FF00)>>(CPQCI_BASE_NORMALPRIORITY_CHANNEL+8);
    if( list )
    {
        //  Found one.  Extract it and reset bit in caller's list.
        *pChannel = (CPQCICHANNEL)
            (Channels[ list ] + (CPQCI_BASE_NORMALPRIORITY_CHANNEL+8));
        *pChannelList &= ~(1 << *pChannel);
        error = CpqCiSuccess;
        goto exitProc;
    }

    DBG_INTR("Check 7:0\n");
    //  Check for normal priority channels 7:0.
    list = (*pChannelList & 0x000000FF)>>CPQCI_BASE_NORMALPRIORITY_CHANNEL;
    if( list )
    {
        //  Found one.  Extract it and reset bit in caller's list.
        *pChannel = (CPQCICHANNEL)
            (Channels[ list ] + CPQCI_BASE_NORMALPRIORITY_CHANNEL);
        *pChannelList &= ~(1 << *pChannel);
        error = CpqCiSuccess;
        goto exitProc;
    }

exitProc:
    return error;
}

/*****************************************************************************
    Private routines.
******************************************************************************/

CPQCIPRIVATE
CPQCIVOID
ConstructKey(
    CPQCIBUFFER pDst,
    CPQCIBUFFER pSrc
    )

/*++

Routine Description:

    This routine derives an access key from the source buffer and
    copies it into the destination buffer.

Arguments:

    pDst - Caller supplied buffer to receive key.

    pSrc - Source buffer used to derive key.

Return Value:

    None.

Caveats:

    This routine uses a known arbitrary pattern to apply to the target
    buffer.  Each pattern byte is combined with the current offset of
    the target buffer plus 1, before being XOR'd with the source byte.

--*/
{
    static u64_t pattern[] = {0x4a2073616d6f6854ULL,0x616c6f6e6f42202eULL};
//    static unsigned long long int pattern[] = {0x4a2073616d6f6854ULL,0x616c6f6e6f42202eULL};
    CPQCIBYTE *psrc = (CPQCIBYTE *)pSrc;
    CPQCIBYTE *pdst = (CPQCIBYTE *)pDst;
    CPQCIBYTE *pdstend = pdst + CPQCI_KEY_SIZE;
    CPQCIBYTE *ppattern = (CPQCIBYTE *)pattern;
    CPQCIBYTE i = 0;

    //  Make sure a destination buffer was supplied.
    if( !pDst )
        goto exitProc;

    //  Computing the key is very simple (maybe too simple).  
    //  We will use an XOR pattern.
    while( pdst < pdstend )
        *pdst++ = *psrc++ ^ (ppattern[i++&0xf]+i);

exitProc:
    return;
}

CPQCIPRIVATE
CPQCISTATUS
MapConnection( 
    CPQCICONTEXT Context,
    CPQCIADAPTER Adapter,
    CPQCICHANNEL Channel,
    CPQCIMODE AddressMode,
    CPQCICCB *pCcb,
    CPQCIHANDLE *pMemoryHandle,
    CPQCIHANDLE *pDoorbellHandle
    )

/*++

Routine Description:

    This routine maps the channel and doorbell memory into the
    caller's virtual address space.

Arguments:

    pDst - Caller supplied buffer to receive key.

    pSrc - Source buffer used to derive key.

Return Value:

    The result of the operation.

Caveats:

    User mode callers will result in handles being created for
    the host memory region and doorbell page.  Kernel mode callers
    will result in the target ccb assuming the driverccb information.

--*/

{
    CPQCIADAPTEROBJECT *pao = Adapter;
    CPQCICCB *driverccb = &pao->ChannelObject[Channel].DriverCcb;
    CPQCIHANDLE mhandle = CPQCINULL;
    CPQCIHANDLE dhandle = CPQCINULL;
    CPQCISTATUS error = CpqCiSuccess;
    CPQCIBYTE *dva;
    CPQCIBYTE *mva;

    /*************************************************************************
        Kernel mode callers get a copy of the driver's ccb addresses.
    **************************************************************************/

    if( AddressMode == CpqCiKernelMode )
    {
        CpqCiCopyBuffer( pCcb, driverccb, sizeof(CPQCICCB) );
        goto exitProc;
    }

    /*************************************************************************
        User mode caller's must map the driver's address information into the
        caller's virtual address space of the current process.
    **************************************************************************/

    //  Map host memory.
    mhandle = 
    CpqCiDrvOsMap( Context,
                   pao->ChannelObject[Channel].MemoryPaAdj.uint,
                   pao->ChannelObject[Channel].MemorySizeAdj,
                   CpqCiUserMode,
                   CpqCiCached  );
    if( !mhandle )
    {
        error = CpqCiErrUnableToMapHostRam;
        goto exitProc;
    }

    //  Map doorbell page aperature.
    dhandle = 
    CpqCiDrvOsMap( Context,
                   pao->ChannelObject[Channel].DbBasePaAdj.uint,
                   CPQCI_PAGE_SIZE,
                   CpqCiUserMode,
                   CpqCiUncached  );

    if( !dhandle )
    {
        error = CpqCiErrUnableToMapDoorbell;
        goto exitProc;
    }

    /*************************************************************************
        Construct the connection's ccb.
    **************************************************************************/

    mva = CpqCiDrvOsGetVA( Context, mhandle );
    dva = CpqCiDrvOsGetVA( Context, dhandle );

    CpqCiCcbSetup( pCcb,
                   Channel,
                   dva,
                   mva + pao->ChannelObject[Channel].SqFifoOffset.uint,
                   mva + pao->ChannelObject[Channel].SqDescOffset.uint,
                   CpqCiCtrlGetDesclimit(driverccb->SqCtrl) + 1,
                   CpqCiDescMemorySize(driverccb->SqCtrl, 1),
                   1,
                   mva + pao->ChannelObject[Channel].RqFifoOffset.uint,
                   mva + pao->ChannelObject[Channel].RqDescOffset.uint,
                   CpqCiCtrlGetDesclimit(driverccb->RqCtrl) + 1,
                   CpqCiDescMemorySize(driverccb->RqCtrl, 1),
                   1  );

    /*************************************************************************
        Exit the routine.
    **************************************************************************/

    error = CpqCiSuccess;

exitProc:
    if( error != CpqCiSuccess )
    {
        CpqCiDrvOsFree( Context, mhandle );
        CpqCiDrvOsFree( Context, dhandle );

        mhandle = 
        dhandle = CPQCINULL;
    }

    if( pMemoryHandle )
        *pMemoryHandle = mhandle;

    if( pDoorbellHandle )
        *pDoorbellHandle = dhandle;

    return error;
}

int CpqCiShareContext(
	CPQCICONTEXT Context,
	struct _CPQCIADAPTEROBJECT *pao,  
	CPQCICONNECTION connection 
	)
{
   	CPQCICONNECTIONOBJECT* connectionObject; 

	if ((connection < 0) || (connection >= pao->MaxConnections)) return -1;

   	connectionObject = &pao->ConnectionObject[connection];
	connectionObject->Context = (CPQCIHANDLE) Context;
	return 0;
}

CPQCIDRVOSHANDLE GetMemoryHandle( 
	struct _CPQCIADAPTEROBJECT *pao,  
	CPQCICONNECTION connection 
	)
{
	if ((connection < 0) || (connection >= pao->MaxConnections)) return NULL;
	return pao->ConnectionObject[connection].MappedMemoryHandle;
}
